/*
     This file is part of GNUnet.
     (C) 2001, 2002 Christian Grothoff (and other contributing authors)

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * ContentDatabase (dispatch implementation, currently bound to GDBM). 
 * @author Christian Grothoff
 * @file applications/afs/database/contentdatabase.c
 **/

#include "config.h"
#include "applications/afs/database/contentdatabase.h"
#ifdef GDBM_STORAGE
#include "applications/afs/database/contentdatabase_gdbm.h"
#elif TDB_STORAGE
#include "applications/afs/database/contentdatabase_tdb.h"
#elif DIRECTORY_STORAGE
#include "applications/afs/database/contentdatabase_directory.h"
#endif
#include "util/xmalloc.h"
#include "util/semaphore.h"
#include "util/configuration.h"

#include <unistd.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <pwd.h>
#include <grp.h>
#include <dirent.h>
#include <errno.h>


void initContentDatabase() {  
  char * dir;

  dir = getFileName("AFS",
		    "DATABASE",
		    "Configuration file must specify filename for the database in section %s under %s.\n");
#ifdef GDBM_STORAGE
  initContentDatabase_GDBM(dir);
#elif TDB_STORAGE
  initContentDatabase_TDB(dir);
#elif DIRECTORY_STORAGE
  initContentDatabase_Directory(dir);
#else
#error YOU MUST HAVE A STORAGE SYSTEM (install tdb or gdbm)
#endif
  xfree(dir,
	"database directory");
}

/**
 * Clean shutdown of the storage module
 **/
void doneContentDatabase() {
#ifdef GDBM_STORAGE
  doneContentDatabase_GDBM();
#elif TDB_STORAGE
  doneContentDatabase_TDB();
#elif DIRECTORY_STORAGE
  doneContentDatabase_Directory();
#endif
}

/**
 * Call a method for each key in the database and
 * call the callback method on it. 
 *
 * @param callback the callback method
 * @param data second argument to all callback calls
 * @return the number of items stored in the content database
 **/
int forEachEntryInDatabase(EntryCallback callback,
			   void * data){
#ifdef GDBM_STORAGE
  return forEachEntryInDatabase_GDBM(callback, data);
#elif TDB_STORAGE
  return forEachEntryInDatabase_TDB(callback, data);
#elif DIRECTORY_STORAGE
  return forEachEntryInDatabase_Directory(callback, data);
#endif
}

/**
 * Get the number of entries in the database.
 * @return SYSERR on error, otherwise the number of entries
 **/
int countContentEntries() {
#ifdef GDBM_STORAGE
  return countContentEntries_GDBM();
#elif TDB_STORAGE
  return countContentEntries_TDB();
#elif DIRECTORY_STORAGE
  return countContentEntries_Directory();
#endif
}
 
/**
 * Read the contents of a bucket to a buffer.
 *
 * @param fn the hashcode representing the entry
 * @param result the buffer to write the result to 
 *        (*result should be NULL, sufficient space is allocated)
 * @return the number of bytes read on success, -1 on failure
 **/ 
int readContent(HashCode160 * name,
		void ** result) {
#ifdef GDBM_STORAGE
  return readContent_GDBM(name, result);
#elif TDB_STORAGE
  return readContent_TDB(name, result);
#elif DIRECTORY_STORAGE
  return readContent_Directory(name, result);
#endif
}

/**
 * Write content to a file. Check for reduncancy and eventually
 * append.
 *
 * @param fn the key for the entry
 * @param len the size of the block
 * @param block the data to store
 * @return SYSERR on error, OK if ok.
 **/
int writeContent(HashCode160 * fn, 
		 int len,
		 void * block) {
#ifdef GDBM_STORAGE
  return writeContent_GDBM(fn, len, block);
#elif TDB_STORAGE
  return writeContent_TDB(fn, len, block);
#elif DIRECTORY_STORAGE
  return writeContent_Directory(fn, len, block);
#endif
}

/**
 * Free space in the database by removing an entry.
 *
 * @param fn the key of the entry to remove
 * @return SYSERR on error, OK if ok.
 **/
int unlinkFromDB(HashCode160 * name) {
#ifdef GDBM_STORAGE
  return unlinkFromDB_GDBM(name);
#elif TDB_STORAGE
  return unlinkFromDB_TDB(name);
#elif DIRECTORY_STORAGE
  return unlinkFromDB_Directory(name);
#endif
}


/* end of contentdatabase.c */
