/*
     This file is part of GNUnet.
     (C) 2001, 2002 Christian Grothoff (and other contributing authors)

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * @file applications/afs/tools/gnunet-check.c
 * @author Christian Grothoff
 *
 * Little tool to do consistency check of the AFS databases.
 **/

#include "config.h"

#include <sys/types.h>
#include <unistd.h>
#include <pwd.h>
#include <pthread.h>
#include <signal.h>
#include <unistd.h>
#include <errno.h>
#include "util/getopt.h"
#include "util/tcp_return.h"
#include "util/tcpio.h"
#include "util/statistics.h"
#include "util/ports.h"
#include "util/hashing.h"
#include "util/initialize.h"
#include "util/logging.h"
#include "applications/afs/afs.h"
#include "applications/afs/util/bloomfilter.h"
#include "applications/afs/util/contentencoding.h"
#include "applications/afs/encoding/insertutil.h"
#include "applications/afs/database/lookup.h"
#include "applications/afs/database/fileindex.h"
#include "applications/afs/database/contentdatabase.h"
#include "applications/common/port.h"
#include "applications/common/printhelp.h"

/**
 * To make the linker happy... (database library links against this symbol, 
 * typically defined in the AFS library, but we don't need to drag that
 * library in).
 **/
void * coreAPI;

/* configuration: do we fix problems? */
static int do_fix = YES;

/* tcp server result: were all the requests satisfied? */
static int tcp_verifies;

static int progressDot = 0;
static int be_verbose = NO;
static int be_quiet = NO;

/**
 * Filters.
 **/
Bloomfilter * superBloomFilter = NULL;
Bloomfilter * singleBloomFilter = NULL;

/**
 * Check that the content at the given offset/file
 * has the given double hash.
 **/
static int checkHashMatch(int fileNameIndex,
			  size_t offset,
			  HashCode160 * chkquery) {
  CONTENT_Block result;
  CONTENT_Block eresult;
  char * fn;
  HashCode160 hc;
  HashCode160 dhc;
  int fileHandle;
  size_t blen;
  /* check that the specified file at the specified
     offset actually contains the content that we
     are looking for */

  fn = getIndexedFileName(fileNameIndex);
  if (fn == NULL)
    return SYSERR;
  fileHandle = open(fn, O_EXCL, S_IRUSR);
  if (fileHandle == -1) {
    LOG(LOG_WARNING, 
	"WARNING: Could not open file %s (%d).\n",
	fn,
	fileNameIndex);    
    xfree(fn, "filename");
    return SYSERR;
  }
  xfree(fn, "filename");
  lseek(fileHandle, 
	offset, SEEK_SET);
  memset(&result, 
	 0, 
	 sizeof(CONTENT_Block));
  blen = read(fileHandle, 
	      &result,
	      sizeof(CONTENT_Block));
  close(fileHandle);
  hash(&result, 
       blen, 
       &hc);
  encryptContent(&result,
		 &hc,
		 &eresult);
  hash(&eresult,
       sizeof(CONTENT_Block),
       &dhc);
  if (!equalsHashCode160(&dhc,
			 chkquery)) 
    return SYSERR;
  else
    return OK;
}

/**
 * Remove an entry from the content database.
 **/
static void removeContent(HashCode160 * doubleHash) {
  /* ok, neither double nor triple hash found; remove content! */
  HexName hn;
  hash2hex(doubleHash,
	   &hn);
  if (be_quiet == NO)
    printf("Content %s not indexed in lookup database. ",
	   (char*) &hn);
  if (do_fix == YES) {	   
    if (SYSERR == unlinkFromDB(doubleHash)) {
      if (be_quiet == NO)	  
	printf("unlinkFromDB returned SYSERR, can not fix!\n");
    } else {
      if (be_quiet == NO)
	printf("Fixed (removed).\n");
    }
  } else
    if (be_quiet == NO)
      printf("\n");
}

/**
 * Check that the lookup database has a mapping for this hashcode and
 * that the file-indices map it to the content database! If there is
 * no correct matching entry in the lookup database, the content is
 * removed from the content database.
 *
 * @param doubleHash the double-hash of the content in the database.
 *   Note that the lookup database may contain the triple-hash for CHK content
 **/
static void indexContent(HashCode160 * doubleHash,
			 void * unused) {
  ContentIndex res;
  CollisionLevel * level;
  HexName hn;

  if (be_verbose == YES) {
    hash2hex(doubleHash,
	     &hn);
    printf("* %s (cdb)\n",
	  (char*)&hn);
  } else {
    progressDot++;
    if ((progressDot & 0xFF) == 0)
      if (be_quiet == NO) {
	printf(".");
	fflush(stdout);
      }
  }
  level = NULL;
  if (NO == findEntry(doubleHash,
		      &res,
		      &level)) {
    /* no double-hash entry found, now try triple hash! */
    HashCode160 tripleHash;

    hash(doubleHash,
	 sizeof(HashCode160),
	 &tripleHash);
    level = NULL;
    if (NO == findEntry(&tripleHash,
			&res,
			&level)) {
      removeContent(doubleHash);
      return;
    } else { 
      /* yes, found with triple-hash query, verify that the
	 entry is ok */
      if ( (equalsHashCode160(&res.hash, doubleHash)) &&
	   (ntohs(res.type) == LOOKUP_TYPE_3HASH) &&
	   (ntohs(res.fileNameIndex) == 0) &&
	   (ntohs(res.fileOffset) == 0) )
	return; /* matching, correct entry found, exit! */
      else {
	removeContent(doubleHash);      
	return;
      }
    }
  } else { 
    /* findEntry succeeded, check if it is what we expected */
    if (ntohs(res.type) == LOOKUP_TYPE_3HASH) {
      LOG(LOG_WARNING,
	  "WARNING: found 3HASH type in lookup database for double-hash lookup, "
	  "this is very unusual.\n");
      removeContent(doubleHash);
      return;
    }
    if (ntohs(res.fileNameIndex) != 0) {
      /* should be zero, this method is only called for
	 content that should be in the contentdatabase;
	 this can happen if we have content both in the
	 content database and on-demand encoded. Either
	 way, get rid of the (potentially duplicated) 
	 content database entry. */
      removeContent(doubleHash);
      return;
    }
  }
}

/**
 * Check that for each entry in the contentdatabase
 * there is an entry in the lookup-database.
 **/
static void checkDatabase() {
  int count;

  if (be_quiet == NO)
    printf("Checking Content Database\n");
  count = forEachEntryInDatabase(&indexContent,
				 NULL);  
  if (be_quiet == NO) {
    if (be_verbose == NO)
      printf("\n");
    printf("==> Done checking %d entries in content database.\n", 
	   count);
  }
}

/**
 * Process a request to insert content from the client.
 * @return SYSERR if the TCP connection should be closed, otherwise OK
 **/
static int checkInsertCHK(GNUNET_TCP_SOCKET * sock,
			  AFS_CS_INSERT_CHK * insertRequest) {
  CONTENT_Block * block;
  int len;
  ContentIndex entry;
  CollisionLevel * level;
  int res;  
  HashCode160 hc;
  HexName hn;

  if (ntohs(insertRequest->header.size) != 
      sizeof(AFS_CS_INSERT_CHK)) {
    sendTCPResult(sock, SYSERR);
    return SYSERR;
  }

  block = NULL;
  hash(&insertRequest->content,
       sizeof(CONTENT_Block),
       &hc);
  hash2hex(&hc,
	   &hn);
  if (be_verbose == YES)
    printf("* %s (ins)\n",
	   (char*)&hn);
  res = SYSERR;
  len = readContent(&hc,
		    (void**)&block);
  if(len == sizeof(CONTENT_Block)) 
    if (0 == memcmp(&insertRequest->content,
  	  	    block,
		    sizeof(CONTENT_Block)) )      
      res = OK;
  xfree(block, "checkInsert: block");
  
  if (res == SYSERR) {
    HexName hn;

    hash2hex(&hc,
	     &hn);
    if (be_quiet == NO)
      printf("Content %s missing in database. ",
	     (char*) &hn);
    if (do_fix == YES) {
      if (OK == writeContent(&hc,
			     sizeof(CONTENT_Block),
			     &insertRequest->content)) {
	if (be_quiet == NO)
	  printf("Fixed.\n");
      } else {
	printf("WARNING: can not fix (database full?)\n");
      }
    } else 
      if (be_quiet == NO)
	printf("\n");
  }      

  /* also check/fix lookup database */
  level = NULL;
  if (NO == findEntry(&hc,
		      &entry,
		      &level)) {
    HexName hn;
    hash2hex(&hc,
	     &hn);
    if (be_quiet == NO)
      printf("Content %s not indexed in lookup database. ",
	     (char*) &hn);
    if (do_fix == YES) {	
      indexContent(&hc, NULL);
      if (be_quiet == NO)
	printf("Fixed.\n");
    } else
      if (be_quiet == NO)
	printf("\n");
  } else {
    if ( (ntohs(entry.fileNameIndex) != 0) ||
	 (ntohl(entry.fileOffset) != 0) ) {
      if (be_quiet == NO)
	printf("Bad value (II) stored in lookup database"
	       "(could be ok) ");
      if (do_fix == YES) {
	indexContent(&hc, NULL);
	if (be_quiet == NO)
	  printf("Fixed.\n");
      } else
	if (be_quiet == NO)
	  printf("\n");
    }
  }
  sendTCPResult(sock, OK);
  return OK;
}

/**
 * Process a request to insert content from the client.
 * @return SYSERR if the TCP connection should be closed, otherwise OK
 **/
static int checkInsert3HASH(GNUNET_TCP_SOCKET * sock,
			    AFS_CS_INSERT_3HASH * insertRequest) {
  LOG(LOG_WARNING,
      "WARNING: did not expect 3HASH insert invocation!\n");
  sendTCPResult(sock, OK);
  return OK;
}

static int checkSuper(GNUNET_TCP_SOCKET * sock,
		      AFS_CS_INDEX_SUPER * superIndexRequest) {
  ContentIndex entry;
  ContentIndex entry2;
  CollisionLevel * level = NULL;

  if (ntohs(superIndexRequest->header.size) != 
      sizeof(AFS_CS_INDEX_SUPER)) {
    LOG(LOG_WARNING, 
	"WARNING: super-hash indexing request from client was malformed!\n");
    return SYSERR;
  }
  if (ntohl(superIndexRequest->isSuper) == YES) {
    if (NO == testBloomfilter(superBloomFilter,
			      &superIndexRequest->superHash)) { 
      if (be_quiet == NO)
	printf("Super-Hash not listed in super-hash bloom filter ");
      if (do_fix == YES) {
	addToBloomfilter(superBloomFilter,
			 &superIndexRequest->superHash);
	if (be_quiet == NO)
	  printf("Fixed.\n");
      } else
	if (be_quiet == NO)
	  printf("\n");
    }
  } else {
    if (NO == testBloomfilter(singleBloomFilter,
			      &superIndexRequest->superHash)) { 
      if (be_quiet == NO)
	printf("Super-Hash not listed in single-hash bloom filter ");
      if (do_fix == YES) {
	addToBloomfilter(singleBloomFilter,
			 &superIndexRequest->superHash);
	if (be_quiet == NO)
	  printf("Fixed.\n");
      } else
 	if (be_quiet == NO)
	  printf("\n");
    }
    return sendTCPResult(sock, 
			 OK); /* content database test is only (!) valid for
				 the real SUPER-hashs, for the others (3HASH), there
				 is a separate request for the content index */
  }
  entry.type
    = htons(LOOKUP_TYPE_SUPER);
  entry.importance
    = 0; /* super-queries are always incredbily important */
  entry.fileNameIndex 
    = 0; /* database */
  entry.fileOffset 
    = 0; /* data/content */
  memcpy(&entry.hash,
	 &superIndexRequest->superHash,
	 sizeof(HashCode160));
  if (SYSERR == findEntry(&superIndexRequest->superHash,
			  &entry2,
			  &level)) {
    /*
      This seems to happen with the release candidate for 0.5.0. The
      problem is completely minor since the super-hashes in the lookup
      database are used at most to repair the bloom-filter when the user
      runs gnunet-check with -i or -a. The super-entries in the lookup
      index are *never* used in normal operation. Anyway, it would be
      interesting to find out why this occurs. A single run of gnunet-check
      -f seems to fix the problem entirely... 
    */
    if (be_quiet == NO) {
      HexName expect;

      hash2hex(&superIndexRequest->superHash,
	       &expect);
      printf("Did not find super-hash entry in "
	     "lookup database for hash %s (known insignificant problem). ",
	     (char*)&expect);
    }
    if (do_fix == YES) {
      addEntry(&entry, NULL);
      if (be_quiet == NO)
	printf("Fixed.\n");
    } else
      if (be_quiet == NO)
	printf("\n"); 
  } else {
    if (0 != memcmp(&entry, &entry2, sizeof(ContentIndex))) {
      if (be_quiet == NO) {
	HexName have;
	HexName expect;

	hash2hex(&entry2.hash,
		 &have);
	hash2hex(&entry.hash,
		 &expect);
	printf("Entry in database for super-hash does not "
	       "match expectations (have: %s, %d, %d, %d; "
	       "expected: %s, %d, %d, %d). ",
	       (char*)&have, entry2.importance,
	       entry2.fileNameIndex, entry2.fileOffset,
	       (char*)&expect, entry.importance, 
	       entry.fileNameIndex, entry.fileOffset);
      }
      if (do_fix == YES) {
	addEntry(&entry, NULL);
	if (be_quiet == NO)
	  printf("Fixed.\n");
      } else
 	if (be_quiet == NO)
	  printf("\n");
    }
  }
  return sendTCPResult(sock, 
		       OK);
}

/**
 * Process a request to index content from the client.
 * @return SYSERR if the TCP connection should be closed, otherwise OK
 **/
static int checkIndex(GNUNET_TCP_SOCKET * sock,
		      AFS_CS_INDEX_BLOCK * indexingRequest) {
  HashCode160 triple;
  HashCode160 * query;
  ContentIndex res;
  CollisionLevel * level;
  HexName hn;

  hash2hex(&indexingRequest->contentIndex.hash,
	   &hn);
  if (be_verbose == YES)
    printf("* %s (idx)\n",
	   (char*)&hn);
  switch (ntohs(indexingRequest->contentIndex.type)) {
  case LOOKUP_TYPE_3HASH:
    hash(&indexingRequest->contentIndex.hash,
	 sizeof(HashCode160),
	 &triple);
    query = &triple;
    break;
  case LOOKUP_TYPE_CHK:
    query = &indexingRequest->contentIndex.hash;
    break;
  default:
    LOG(LOG_ERROR,
	"ERROR: Unexpected content index type: %d\n",
	ntohs(indexingRequest->contentIndex.type));
    return SYSERR; 
  }
  if (ntohs(indexingRequest->header.size) != 
      sizeof(AFS_CS_INDEX_BLOCK)) {
#if PRINT_TCP
    printf("TCP: WARNING: indexing request malformed!\n");
#endif
    sendTCPResult(sock, SYSERR);
    return SYSERR;
  }
#if PRINT_TCP 
  printf("TCP: received indexing request\n");
#endif
  /* check if everything is already in place, and if not and 
     we are allowed to fix, do the write-actions: */

  level = NULL;
  if (NO == findEntry(query,
		      &res,
		      &level)) {
    HexName hn;
    hash2hex(query,
	     &hn);
    if (be_quiet == NO)
      printf("Content %s not indexed in lookup database. ",
	     (char*) &hn);
    if (do_fix == YES) {	   
      addEntry(&indexingRequest->contentIndex, NULL);
      if (be_quiet == NO)
	printf("Fixed.\n");
    } else
      if (be_quiet == NO)
	printf("\n");
  } else {
    if (0 != memcmp(&res.hash, 
		    &indexingRequest->contentIndex.hash,
		    sizeof(HashCode160))) {
      if (be_quiet == NO)
	printf("Bad value (IR) stored in lookup database (could be ok) ");
      if (do_fix == YES) {
	addEntry(&indexingRequest->contentIndex, NULL);
	if (be_quiet == NO)
	  printf("Fixed.\n");
      } else
	if (be_quiet == NO)
	  printf("\n");
    }
  }
  sendTCPResult(sock, OK);
  return OK;
}


/**
 * Process a query to list a file as on-demand encoded from the client.
 * (code copied from afs/handler.c).
 *
 * @return SYSERR if the TCP connection should be closed, otherwise OK
 **/
static int csHandleRequestIndexFile(GNUNET_TCP_SOCKET * sock,
				    AFS_CS_INDEX_FILE * listFileRequest) {
  if (ntohs(listFileRequest->header.size) != 
      sizeof(AFS_CS_INDEX_FILE)) {
    LOG(LOG_WARNING, 
	"WARNING: file indexing request from client was malformed!\n");
    return SYSERR;
  }
  listFileRequest->filename[1023] = 0; /* ensure strlen terminates! */
  return sendTCPResult(sock, 
		       appendFilename(listFileRequest->filename));
}


/**
 * Handle data available on the TCP socket descriptor;
 * check that the request is already fullfilled.
 **/
static void checkProcessor(int sockDescriptor){
  CS_HEADER * hdr;
  GNUNET_TCP_SOCKET sock;
  int i;

  /* register the socket */
  initGNUnetServerSocket(sockDescriptor,
			 &sock);
  while (1) {
    hdr = NULL;    
    if (SYSERR == readFromSocket(&sock,
				 &hdr) )
      break; /* connection closed */
    /* demultiplex */
    switch (ntohs(hdr->tcpType)) {
    case AFS_CS_PROTO_INDEX_FILE:
      i = csHandleRequestIndexFile(&sock,
				   (AFS_CS_INDEX_FILE*)hdr);
      break;
    case AFS_CS_PROTO_INSERT_3HASH:
      i = checkInsert3HASH(&sock,
			   (AFS_CS_INSERT_3HASH*)hdr);
      break;
    case AFS_CS_PROTO_INSERT_CHK:
      i = checkInsertCHK(&sock,
			 (AFS_CS_INSERT_CHK*)hdr);
      break;
    case AFS_CS_PROTO_INDEX_BLOCK:
      i = checkIndex(&sock,
		     (AFS_CS_INDEX_BLOCK*) hdr);
      break;
    case AFS_CS_PROTO_INDEX_SUPER: 
      i = checkSuper(&sock,
		     (AFS_CS_INDEX_SUPER*) hdr);
      break;
    default:
      i = SYSERR;
      break;
    } /* end of switch */
    if (OK != i) {
      break;
    }
    xfree(hdr,
	  "freeing TCP buffer");
  }
  destroySocket(&sock);
}

/**
 * Check that the given file is properly indexed
 * (and fix if appropriate). Also return SYSERR
 * if the file is gone and should thus be removed
 * from the list.
 **/
int checkIndexedFile(char * name,
		     int index,
		     GNUNET_TCP_SOCKET * sock) {
  int result;
  Block * top;

  if (be_quiet == NO)
    printf("* %s\n",
	   name);
  top = insertFile(sock,
		   name,
		   NULL,
		   NULL);
  if (top != NULL) {
    top->done(top, NULL);
    result = tcp_verifies;
  } else
    result = SYSERR;
  if (result == SYSERR) {
    if (be_quiet == NO)
      printf("Problem checking indexing of file %s ",
	     name);
    if (do_fix == YES) {
      if (be_quiet == NO)
	printf("Removing file from list.\n");
      return SYSERR; /* remove file, there was a problem */
    } else {
      if (be_quiet == NO)
	printf("\n");
      return OK;
    }
  }
  return OK;
}

/**
 * Check that all files that are listed in
 * the list of indexed files actually exist
 * and that they are properly indexed in the
 * lookup (triple->double hash) database.
 **/
static void checkIndexedFileList() {
  GNUNET_TCP_SOCKET * sock;
  int count;

  sock = getClientSocket();
  if (be_quiet == NO)
    printf("Checking indexed files\n");
  count = forEachIndexedFile((IndexedFileNameCallback)&checkIndexedFile,
			     sock);
  if (be_quiet == NO)
    printf("==> Done with %d indexed files.\n",
	   count);
  releaseClientSocket(sock);
}

/**
 * Print a list of the options we offer.
 **/
static void printhelp() {
  printf("USAGE: gnunet-check [OPTIONS]\n"
	 "-p PRIORITY, --prio=PRIORITY  : "
	 "priority of the inserted content (default: %d)\n"
	 "-r, --reset                   : "
	 "reset bloom-filter (includes 'a' option)\n"
	 "-n, --nofix                   : "
	 "do not fix problems\n"
	 "-a, --all                     : "
	 "check all three databases\n"
	 "-i, --index                   : "
	 "only check the index database\n"
	 "-D, --data                    : "
	 "only check the content database\n"
	 "-f, --files                   : "
	 "only check the file-list\n"
	 "-V, --verbose                 : "
	 "be verbose\n"
	 "-q, --quiet                   : "
	 "be quiet\n",
	 getConfigurationInt("GNUNET-INSERT",
			     "CONTENT-PRIORITY"));
  printHelpDefault();
  printf("Never run gnunet-check while gnunetd is running!\n");
}

/**
 * Perform option parsing from the command line. 
 **/
static int parseCommandLine(int argc, 
			    char * argv[]) {
  int shouldexit=0;
  int c;

  freeNonNull(setConfigurationString("GNUNETD",
				     "LOGFILE",
				     NULL),
	      "gnunet-check always use option d");
  while (1) {
    int option_index = 0;
    static struct option long_options[] = {
      LONG_DEFAULT_OPTIONS,
      { "nofix",   0, 0, 'n' },
      { "index",   0, 0, 'i' },
      { "data",    0, 0, 'D' },
      { "prio",    1, 0, 'p' },
      { "files",   0, 0, 'f' },
      { "all",     0, 0, 'a' },
      { "verbose", 0, 0, 'V' },
      { "quiet",   0, 0, 'q' },
      { 0,0,0,0 }
    };
    
    c = getopt_long(argc,
		    argv, 
		    "vhdc:nDp:faiVqL:", 
		    long_options, 
		    &option_index);    
    if (c == -1) 
      break;  // No more flags to process
    if (YES == parseDefaultOptions(c, optarg))
      continue;    
    switch(c) {
    case 'L':
      freeNonNull(setConfigurationString("GNUNETD",
					 "LOGLEVEL",
					 optarg),
		  "gnunet-check option parsing L");
     break;
    case 'q':
      be_quiet = YES;
      break;
    case 'a':
      freeNonNull(setConfigurationString("GNUNET-CHECK",
					 "MODE",
					 "a"),
		  "gnunet-check option parsing a");
      break;
    case 'i':
      freeNonNull(setConfigurationString("GNUNET-CHECK",
					 "MODE",
					 "i"),
		  "gnunet-check option parsing i");
      break;
    case 'f':
      freeNonNull(setConfigurationString("GNUNET-CHECK",
					 "MODE",
					 "f"),
		  "gnunet-check option parsing f");
      break;
    case 'V':
      be_verbose = YES;
      break;
    case 'D':
      freeNonNull(setConfigurationString("GNUNET-CHECK",
					 "MODE",
					 "d"),
		  "gnunet-check option parsing d");
      break;
    case 'p': {
      unsigned int contentPriority;
      
      if (1 != sscanf(optarg, "%ud", &contentPriority)) {
	LOG(LOG_FAILURE,
	    "FAILURE: You must pass a number to the -p option.\n");
	return -1;
      }
      setConfigurationInt("GNUNET-INSERT",
			  "CONTENT-PRIORITY",
			  contentPriority);
      break;
    }
    case 'n':
     do_fix = NO;
      break;
    case 'v': 
      printf("GNUnet v%s, gnunet-check v%s\n",
	     VERSION, AFS_VERSION);
      shouldexit=1;
      break;
    case 'h': 
      printhelp(); 
      shouldexit=1;
      break;
    default:
      printf("Unknown option %c. Aborting.\nUse --help to get a list of options.\n",
	    c);
      shouldexit=1;    
    } /* end of parsing commandline */
  }
  if (optind < argc) {
    printf("Invalid arguments: ");
    while (optind < argc)
      printf("%s ", argv[optind++]);
    printf("\nExiting.\n");
    return 1;
  }
  if ( (do_fix == NO) &&
       (be_quiet == NO) )
     printf("You selected verification only, will not fix problems!\n");
  readConfiguration();
  return shouldexit;
}

/**
 * At the given position there is an entry that is
 * garbage, fix if we are allowed to.
 **/
static void deleteGarbage(CollisionLevel * level,
			  size_t pos) {
  int noll = 0;
  int i;

  if (do_fix == YES) {
    lseek(level->handle,
	  pos,
	  SEEK_SET);
    for (i=0;i<sizeof(ContentIndex)/sizeof(int);i++)
      write(level->handle, &noll, sizeof(int));
    if (be_quiet == NO)
      printf("Fixed\n");
  } else {
    if (be_quiet == NO)
      printf("\n");
  }
}

/**
 * Verify that the given entry from the 
 * triple-double-hash database is correct.
 **/
static int verifyEntryOk(ContentIndex * entry,
			 CollisionLevel * level,
			 size_t pos,
			 void * data) {
  size_t expectPos;
  HashCode160 tripleHash;
  HashCode160 * query;
  HexName hn;

  progressDot++;
  if ((progressDot & 255) == 0)
    if (be_quiet == NO) {
      printf(".");
      fflush(stdout);
    }
  if (ntohs(entry->type) == LOOKUP_TYPE_3HASH) {
    hash(&entry->hash,
	 sizeof(HashCode160),
	 &tripleHash);
    query = &tripleHash;
  } else
    query = &entry->hash;
  hash2hex(query,
	   &hn);
  if (be_verbose == YES)
    printf("* %s (idb)\n",
	   (char*)&hn);
  expectPos = (((int*)query)[level->offset] & level->mask)*sizeof(ContentIndex);
  if ((pos - expectPos) > (LOOKUP_CORUN * sizeof(ContentIndex))) {
    if (be_quiet == NO)
      printf("Entry %s in database is not at expected position (%d - %d > %d) ",
	     (char*)&hn, 
	     pos,
	     expectPos,
	     LOOKUP_CORUN * sizeof(ContentIndex));
    deleteGarbage(level, pos);
    return OK;
  }
  if (ntohs(entry->type) == LOOKUP_TYPE_SUPER)
    return OK; /* no real content that corresponds to super-hashes */

  /* make sure that every item that occurs in the
     (triple-hash2double-hash) index is actually reachable and makes
     sense. */
  if (ntohs(entry->fileNameIndex) == 0) {
    int res;
    void * block;

    res = readContent(&entry->hash,
		      &block);
    if (res == SYSERR) {
      if (be_quiet == NO)
	printf("Inserted content %s-%d not found by readContent (%d). ",
	       (char*) &hn, 
	       ntohs(entry->type),
	       res);
      deleteGarbage(level, pos);
      return OK;
    }
  } else {
    char * fn;
    fn = getIndexedFileName(ntohs(entry->fileNameIndex));
    if (fn == NULL) {
      if (be_quiet == NO)
	printf("Bad index into file list: %d ",
	       ntohs(entry->fileNameIndex));
      deleteGarbage(level, pos);
      return OK;      
    }
    if (SYSERR == checkHashMatch(ntohs(entry->fileNameIndex),
				 ntohl(entry->fileOffset),
				 &entry->hash)) {
      printf("Indexed content at offset %d in file %s has changed. ",
	     ntohl(entry->fileOffset), 
	     fn);
      deleteGarbage(level, pos);
      xfree(fn, "filename");
      return OK;      
    }         
    xfree(fn, "filename");
  }
  return OK;
}

/**
 * Verify that the triple-hash to double-hash mapping
 * database (lookup) is consistent.
 **/
static void checkIndexDatabase() {
  int count;

  if (be_quiet == NO)
    printf("Checking index database\n");
  count = forEachLookupEntry(&verifyEntryOk,
			     NULL);
  if (be_quiet == NO) {
    if (be_verbose == NO)
      printf("\n");
    printf("==> Done checking %d entries in index (usage: %d%%).\n",
	   count, 
	   (int) (count*100.0/CACHE_HASH_TABLE_SIZE));
  }
}

static Semaphore * serverSignal;

/**
 * Initialize the TCP port and listen for incoming connections.
 **/
static void tcpListenMain() {
  int listenerFD;
  int incomingFD;
  int lenOfIncomingAddr;
  int listenerPort;
  struct sockaddr_in serverAddr, clientAddr;
  pthread_t incomingThread; 	/* thread handling a new connection */
  const int on = 1;

  listenerPort = getGNUnetPort(); 
  /* create the socket */
  if ( (listenerFD = socket(PF_INET, SOCK_STREAM, 0)) < 0) 
    errexit("Error opening socket. Is gnunetd running?\n");
 
  /* fill in the inet address structure */
  memset((char *) &serverAddr, 0, sizeof(serverAddr));
  serverAddr.sin_family 
    = AF_INET;
  serverAddr.sin_addr.s_addr
    = htonl(INADDR_ANY);
  serverAddr.sin_port   
    = htons(listenerPort);
 
  if ( setsockopt(listenerFD, 
		  SOL_SOCKET, 
		  SO_REUSEADDR, 
		  &on, sizeof(on)) < 0 )
    perror("setsockopt");

  /*bind the socket */
  if (bind (listenerFD, 
	    (struct sockaddr *) &serverAddr,
	    sizeof(serverAddr)) < 0)
    errexit("Error (%s) binding the TCP listener to port. Is gnunetd running?\n",
	    strerror(errno));
  
  /* start listening for new connections */
  listen(listenerFD, 1); 
  semaphore_up(serverSignal);
  /* process incoming data */
  while (1) {
    /* wait for a connection and process it */
    lenOfIncomingAddr = sizeof(clientAddr);
    incomingFD = accept(listenerFD,
			(struct sockaddr *)&clientAddr, 
			&lenOfIncomingAddr);
    if(incomingFD < 0){
      LOG(LOG_ERROR, 
	  "ERROR accepting new connection.\n");
      continue;
    }
    LOG(LOG_DEBUG, 
	"TCP: starting server\n");
    if ((PTHREAD_CREATE(&incomingThread, 
			(void * (*)()) &checkProcessor, 
			(void *) incomingFD,
			16*1024)) != 0) {
      LOG(LOG_ERROR, 
	  "Error creating thread to handle new incoming connection.\n");    
      close(incomingFD);
    } else {
      pthread_detach(incomingThread);
    }
  } /* while (1) */
} 

int main(int argc, char * argv[]) {
  pthread_t tcpPseudoServer;
  char * checkString;
  char check;
  
  if (SYSERR == initUtil(argc, argv, &parseCommandLine))
    return 0;

  checkString = getConfigurationString("GNUNET-CHECK", 
				       "MODE");
  if (checkString != NULL)
    check = checkString[0];
  else
    check = 'n';
  if (check == 'n') {
    fprintf(stderr,
	    "You must choose what to check (specify -D, -f, -i or -a).\n");
    return -1;
  }
  initContentDatabase();
  initFileIndex();
  initLookup();

  /* Reset and fix bloomfilters only if everything is checked */
  if ( (do_fix == YES) && 
       ( (check == 'a') ||
	 (check == 'f') ) ) {
    char * fn;
    fn = getFileName("AFS",
		     "BLOOMFILTER1",
		     "You must specify a filename for the bloomfilter"
		     "in section %s under %s");
    superBloomFilter
       = loadBloomfilter(fn,
                 SUPERBLOOMFILTER_SIZE,
                 SUPERBLOOMFILTER_K);
    xfree(fn, "freeing bloomfilter1 filename");
    fn = getFileName("AFS",
		     "BLOOMFILTER2",
		     "You must specify a filename for the"
		     "bloomfilter in section %s under %s");
    singleBloomFilter
      = loadBloomfilter(fn,
		SINGLEBLOOMFILTER_SIZE,
  		SINGLEBLOOMFILTER_K);
    xfree(fn, 
	  "freeing bloomfilter2 filename");
  }

  serverSignal = new_semaphore(0);
  if (0 != PTHREAD_CREATE(&tcpPseudoServer,
			  (void * (*)()) &tcpListenMain, 
			  NULL,
			  16*1024))
    errexit("FATAL: Could not create tcpServer thread\n");
  semaphore_down(serverSignal);
  free_semaphore(serverSignal);
  if (0 != pthread_detach(tcpPseudoServer))
    LOG(LOG_WARNING,
	"WARNING: could not detach tcpServer thread\n");
 
  if ((check == 'a') || (check == 'd'))
    checkDatabase(); 
  if ((check == 'a') || (check == 'f'))
    checkIndexedFileList();
  if ((check == 'a') || (check == 'i'))
    checkIndexDatabase();

  freeBloomfilter(superBloomFilter);
  freeBloomfilter(singleBloomFilter);
  doneLookup();
  doneFileIndex();
  doneUtil();
  return 0;
}


/* end of check_database.c */
