/*
     This file is part of GNUnet.
     (C) 2001, 2002 Christian Grothoff (and other contributing authors)

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * Main function to obtain statistics from GNUnet.
 *
 * @author Christian Grothoff
 * @author Igor Wronsky
 * @file applications/stats/gnunet-stats.c 
 **/

#include <unistd.h>
#include "config.h"
#include "applications/common/port.h"
#include "applications/common/printhelp.h"
#include "util/getopt.h"
#include "util/statistics.h"
#include "util/configuration.h"
#include "util/logging.h"
#include "util/checksum.h"
#include "util/initialize.h"

#define STATS_VERSION "2.0.0"

/**
 * Print statistics received from TCP socket.
 * @param stream where to print the statistics
 * @param sock the socket to use 
 * @return OK on success, SYSERR on error
 **/
static int requestAndPrintStatistics(FILE * stream,
				     GNUNET_TCP_SOCKET * sock) {
  STATS_CS_MESSAGE * statMsg;
  CS_HEADER csHdr;
  int count;
  int i;
  int mpos;
  
  csHdr.size 
    = htons(sizeof(CS_HEADER));
  csHdr.tcpType
    = htons(STATS_CS_PROTO_GET_STATISTICS);
  if (SYSERR == writeToSocket(sock,
			      &csHdr)) {
    fprintf(stream,
	    "Error sending request for statistics to gnunetd.\n");
    return SYSERR;
  }
  statMsg 
    = xmalloc(MAX_BUFFER_SIZE,
	      "allocating space for tcp buffer");
  statMsg->totalCounters 
    = htonl(1); /* to ensure we enter the loop */
  count = 0;
  while ( count < ntohl(statMsg->totalCounters) ) {
    if (SYSERR == readFromSocket(sock,
				 (CS_HEADER**)&statMsg)) {
      fprintf(stream,
	      "Error receiving reply for statistics from gnunetd.\n");
      xfree(statMsg,
	    "freeing space from tcp buffer");
      return SYSERR;    
    }
    if (ntohs(statMsg->header.size) < sizeof(STATS_CS_MESSAGE)) {
      LOG(LOG_WARNING,
	  "WARNING: received malformed stats message\n");
      break;
    }
    mpos = sizeof(unsigned long long) * ntohl(statMsg->statCounters);
    if (count == 0) {
      fprintf(stream,
	      "%-60s: %16u\n",
	      "Uptime (seconds)",
	      (unsigned int) 
	      ((cronTime(NULL) - ntohll(statMsg->startTime))/cronSECONDS));
    }
    for (i=0;i<ntohl(statMsg->statCounters);i++) {
      if (mpos+strlen(&statMsg->descriptions[mpos])+1 > 
	  ntohs(statMsg->header.size) - sizeof(STATS_CS_MESSAGE)) {
	LOG(LOG_WARNING,
	    "WARNING: received malformed stats message\n");
	break; /* out of bounds! */      
      }
      fprintf(stream,
	      "%-60s: %16llu\n",
	      &statMsg->descriptions[mpos],
	      ntohll(statMsg->values[i]));
      mpos += strlen(&statMsg->descriptions[mpos])+1;
    }    
    count += ntohl(statMsg->statCounters);
  } /* end while */
  xfree(statMsg,
	"freeing space from tcp buffer");
  return OK;
}

/**
 * Prints the usage information for this command if the user errs.
 * Aborts the program.
 **/
static void printhelp() {
  printf("USAGE: gnunet-stats [options]\n");
  printHelpDefault();
}

/**
 * Parse the options, set the timeout.
 * @param argc the number of options
 * @param argv the option list (including keywords)
 * @param keywords the hashs of the keywords to set (= the keys, not the queries!)
 * @return SYSERR if we should abort, OK to continue
 **/
static int parseOptions(int argc,
			char ** argv) {
  int option_index;
  int c;  

  while (1) {
    static struct option long_options[] = {
      LONG_DEFAULT_OPTIONS,
      { 0,0,0,0 }
    };    
    option_index = 0;
    c = getopt_long(argc,
		    argv, 
		    "vhdc:L:H:", 
		    long_options, 
		    &option_index);    
    if (c == -1) 
      break;  // No more flags to process
    if (YES == parseDefaultOptions(c, optarg))
      continue;
    switch(c) {
    case 'v': 
      printf("GNUnet v%s, gnunet-stats v%s\n",
	     VERSION, STATS_VERSION);
      return SYSERR;
    case 'h': 
      printhelp(); 
      return SYSERR;
    default: 
      LOG(LOG_FAILURE,
	  "Unknown option %c. Aborting.\n"\
	  "Use --help to get a list of options.\n",
	  c);
      return -1;
    } /* end of parsing commandline */
  } /* while (1) */
  return OK;
}

/**
 * The main function to obtain statistics from gnunetd.
 *
 * @param argc number of arguments from the command line
 * @param argv command line arguments
 * @return 0 ok, 1 on error
 **/   
int main(int argc, char ** argv) {
  int res;
  GNUNET_TCP_SOCKET * sock;

  if (SYSERR == initUtil(argc, argv, &parseOptions))
    return 0;
  sock = getClientSocket();
  res = requestAndPrintStatistics(stdout,
				  sock);
  releaseClientSocket(sock);
  doneUtil();

  return (res == OK) ? 0 : 1;
}

/* end of gnunet-stats.c */
