/*
     This file is part of GNUnet

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/
/**
 * ContentDatabse - management of the blocks...
 *
 * @author Christian Grothoff
 * @file include/applications/afs/database/contentdatabase.h
 **/

#ifndef CONTENTDATABASE_H
#define CONTENTDATABASE_H

#include "config.h"
#include "util/hashing.h"

/**
 * Signature of a callback method that can be invoked
 * for every key in the database.
 *
 * @param key the key in the database
 * @param data an opaque pointer to pass a context to the callee
 **/
typedef void (*EntryCallback)(HashCode160 * key,
			      void * data);

/**
 * Initialize the storage module.
 *
 * @param dir the name of the directory/file
 *        containing the content database
 **/
void initContentDatabase();

/**
 * Clean shutdown of the storage module.
 **/
void doneContentDatabase();

/**
 * Free space in the database by removing an entry.
 *
 * @param fn the key of the entry to remove
 * @return SYSERR on error, OK if ok.
 **/
int unlinkFromDB(HashCode160 * fn);

/**
 * Get the number of entries in the database.
 * @return SYSERR on error, otherwise the number of entries
 **/
int countContentEntries();
 
/**
 * Read the contents of a bucket to a buffer.
 *
 * @param fn the hashcode representing the entry
 * @param result the buffer to write the result to 
 *        (*result should be NULL, sufficient space is allocated)
 * @return the number of bytes read on success, SYSERR on failure
 **/ 
int readContent(HashCode160 * fn,
		void ** result);

/**
 * Write content to a file. Check for reduncancy and eventually
 * append.
 *
 * @param fn the key for the entry
 * @param len the size of the block
 * @param block the data to store
 * @return SYSERR on error, OK if ok.
 **/
int writeContent(HashCode160 * fn, 
		 int len,
		 void * block);

/**
 * Call a method for each key in the database and
 * call the callback method on it. 
 *
 * @param callback the callback method
 * @param data second argument to all callback calls
 * @return the number of items stored in the content database
 **/
int forEachEntryInDatabase(EntryCallback callback,
			   void * data);
 
#endif

/* end of contentdatabase.h */
