/*
     This file is part of GNUnet

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * @file include/applications/afs/database/lookup.h
 * @author Christian Grothoff
 *
 * This module is responsible for locating available content.
 * After a query is received, the bloomfilter is considered
 * to test if the content is likely to be available. If the
 * bloomfilter-test succeeds, the lookup module is invoked
 * to test if the content is there, and if, tell where it
 * is stored. For 3HASH-queries, the lookup module also 
 * produces the double-hash (inverts the hash).
 **/

#ifndef LOOKUP_H
#define LOOKUP_H

#include "config.h"
#include "util/hashing.h"

/* **************** types ************* */

#define LOOKUP_TYPE_FREE 0
#define LOOKUP_TYPE_DELETED 1
#define LOOKUP_TYPE_CHK 2
#define LOOKUP_TYPE_3HASH 3
#define LOOKUP_TYPE_SUPER 4

/**
 * Type of the content index file entries. The size of this
 * struct dominates the database size, so keep it as small
 * as possible. 32 byte should be enough!
 **/
typedef struct {
  /**
   * The double-hash (hash of the hash of the plaintext) of this entry
   * for 3HASH entries, or the CHK query hash (hash of the encrypted
   * content) for CHK entries. Which is the case can be determined by
   * looking at fileNameIndex and fileOffset.
   **/
  HashCode160 hash;

  /**
   * The current rating of this content (in network byte order).
   **/
  unsigned int importance;

  /**
   * The type of the entry. See LOOKUP_TYPE_XXX
   *
   * The field is always in network byte order.
   **/
  unsigned short type;

  /**
   * This field gives the index of the file into the
   * file-index module if the value is >0. If the
   * value is 0, the file is in the contentdatabase.
   *
   * The field is always in network byte order.
   **/
  unsigned short fileNameIndex; 

  /** 
   * The offset in the file for on-demand-encoded files
   * where fileNameIndex is >0.
   *
   * The field is always in network byte order.
   **/
  size_t fileOffset;  
} ContentIndex;

/**
 * The hashtable consists of multiple levels, a collision in level n
 * is eventually spilled over to level n+1 (after searching CORUN next
 * slots). The number of levels is LOOKUP_LEVELS.
 **/
typedef struct CollisionLevel {
  /**
   * Handle to the collision file for this level 
   **/
  int handle;

  /**
   * Size of this collision table (bit-mask) 
   **/
  int mask;

  /**
   * Offset into the hashcode to mask out (0=.a, 4=.b, 8=.c, etc). 
   **/
  int offset;

  /**
   * Mutex for access to the file.
   **/
  Mutex lock;

  /**
   * The next collision level, linked list is NULL terminated
   **/
  struct CollisionLevel * next;
} CollisionLevel;

/**
 * Initialize the lookup module.
 **/
void initLookup();

/**
 * Shutdown the lookup module.
 **/
void doneLookup();

/**
 * Add an entry to the cache. An existing entry with the same
 * hash will be overwritten.
 *
 * @param entry the entry to add
 * @param level the level at which the entry is located,
 *        pass NULL if the level is not known or the entry
 *        is new.
 * @return OK on success, SYSERR on error (typical errors
 *   are that the database is full and that we had collisions
 *   throughout all levels of the DB; this should happen if
 *   we index more than CACHE_HASH_TABLE_SIZE blocks of data)
 **/
int addEntry(ContentIndex * entry,
	     CollisionLevel * level);

/**
 * Delete an entry from the cache.
 *
 * @param entry the entry to delete
 * @param level the level at which the entry is located,
 *        pass NULL if the level is not known or the entry
 *        is new.
 * @return OK on success, SYSERR on error (errors are not
 *        found or not implemented)
 **/
int delEntry(ContentIndex * entry,
	     CollisionLevel * level);

/**
 * Lookup the given Hash in the lookup database structure
 * (load blocks as needed). 
 *
 * @param query the query to look for, can be a CHK or 3HASH
 *        query.
 * @param result where to write the lookup information on
 *        success
 * @param level is set to the colevel where the lookup was
 *        successful. This can be used to speed up an update.
 * @returns OK if found, SYSERR if not
 **/
int findEntry(HashCode160 * query,
	      ContentIndex * result,
	      CollisionLevel ** level);
 
/**
 * Callback method for iterating over the lookup database.
 *
 * @param entry the entry in the database
 * @param level the collision level at which the entry was found
 * @param pos the position in the file at which the entry resides
 * @param data context for the callee
 * @return SYSERR to abort the iterator, OK to continue
 **/
typedef int (*ContentIndexCallback)(ContentIndex * entry,
				    CollisionLevel * level,
				    size_t pos,
				    void * data);

/**
 * Call a method for each entry in the lookup table.
 * VERY expensive (will take quite a while to run),
 * so be careful when using this.
 *
 * @param method the method to call, abort early if 
 *        method returns SYSERR
 * @param data last argument to the callback method
 * @return the number of entries in the database
 **/
int forEachLookupEntry(ContentIndexCallback method,
		       void * data);

/**
 * Iterator over the database.
 **/
typedef struct LookupDatabaseIterator{

  /**
   * Internal state of the iterator
   **/ 
  CollisionLevel * current;
  unsigned int pos;
  ContentIndex entry;

  /**
   * How many elements has this iterator covered so far?
   **/
  unsigned int count;
  
} LookupDatabaseIterator;

/**
 * Create a lookup database Iterator.
 **/
LookupDatabaseIterator * createLDI();

/**
 * Destroy an LDI.
 **/
void destroyLDI(LookupDatabaseIterator * ldi);

/**
 * Is there a next element in the database?
 * Always call before next!
 * @return YES if there is, NO if not.
 **/
int ldiHasNext(LookupDatabaseIterator * ldi);
   
/**
 * Call the callback method on the next element.
 **/
void ldiGetNext(LookupDatabaseIterator * ldi,
		ContentIndex * entry,
		CollisionLevel ** level);

/**
 * Number of Entries in the on-disk database. Should be larger than
 * the number of 1k blocks shared and must be a power of 2. The
 * database-file will be 32bytes * CACHE_HASH_TABLE_SIZE. If the table
 * is sparse, the database-file may occupy less actual space on the
 * drive. Default is 8388608, which corresponds to a 256 MB database
 * file.  This is sufficient to share about 8 GB of data (yet, if you
 * get closer to this number, efficiency degrades and you may want to
 * double the size of this table). You can increase this number up to
 * 2^31 (= 2 GB database, or 64 GB shared files). Note that gnunetd
 * tends to access this file in a random fashion and thus the kernel
 * will like to buffer the entire file in memory (which means that you
 * will only get good performance if you have more memory than the
 * size of this file). After you change this value, you must run
 * "gnunet-check -a". Expect some loss of information (in particular
 * which content is important) and gnunet-check running for a long,
 * long time, though.  If you dedicate more space to GNUnet, you may
 * want to consider running multiple nodes instead.<p>
 *  
 * The space of this database was used to only 6% in GNUnet 0.4.5
 * until we got into trouble, the new format will stop being able to
 * store data at about 94% of the database being full. See also
 * src/historical/simula.c for simulation code. 
 **/
#define CACHE_HASH_TABLE_SIZE 8388608  

/**
 * Definitions for lookup.c; see also: src/historical/simula.c 
 **/
#define LOOKUP_LEVELS 3
#define LOOKUP_CORUN 16
#define LOOKUP_POWER_FACTOR 2 


#endif
/* end of lookup.h */
