/*
     This file is part of GNUnet

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/** 
 * @file include/applications/afs/encoding/requestmanager.h
 * @author Christian Grothoff
 *
 * The RequestManager keeps track of queries and re-issues requests if
 * no reply is received.
 **/ 

#ifndef REQUESTMANAGER_H
#define REQUESTMANAGER_H

#include "config.h"
#include "util/semaphore.h"
#include "util/cron.h"
#include "util/tcpio.h"
#include "server/coreapi.h"

#include "applications/afs/protocol/messages.h"
#include "applications/afs/encoding/block.h"

/**
 * Create a request manager. Will create the request manager
 * datastructures and also connect to gnunetd. Creates thread that
 * listens to gnunetd replies and another thread that periodically
 * re-issues the queries. Use destroyRequestManager to abort and/or to
 * free resources after the download is complete. The callback method
 * in nc will be invoked to notify the caller of the download progress
 * such that it is possible to tell when we are done.
 *
 * @return NULL on error
 **/
RequestManager * createRequestManager();

/**
 * Destroy the resources associated with a request manager.
 * Invoke this method to abort the download or to clean up
 * after the download is complete.
 *
 * @param this the request manager struct from createRequestManager
 **/
void destroyRequestManager(RequestManager * this);

/**
 * This method receives data corresponding to the indicated filename
 * (hashcode). Finds the Listener that scheduled this request and drop
 * it from the list of pending requests.<p>
 * 
 * @param this the request manager struct from createRequestManager
 * @param msg the message received from gnunetd
 **/ 
void requestManagerReceive(RequestManager * this,
			   AFS_CS_RESULT_CHK * msg);

/**
 * For debugging.
 **/
void printRequestManager(RequestManager * this);

/**
 * Assert that there are no pending requests for this node.
 **/
void requestManagerAssertDead(RequestManager * this,
			      Block * node);

/**
 * Queue a request for execution.
 * 
 * @param this the request manager struct from createRequestManager
 * @param node the node to call once a reply is received
 * @param callback the method to invoke
 * @param data the data argument to the Listener
 * @param message the query to send to gnunetd, freed by callee!
 **/
void requestManagerRequest(RequestManager * this,
			   Block * node,
			   Listener callback,
			   void * data,
			   AFS_CS_QUERY * message);


/**
 * Update a request. This method is used to selectively
 * change a query or drop it entirely.
 *
 * @param this the request manager struct from createRequestManager
 * @param node the block for which the request is updated
 * @param msg the new query message for that node, NULL for 
 *        none (then the request is dropped)
 **/
void requestManagerUpdate(RequestManager * this,
			  Block * node,
			  AFS_CS_QUERY * msg);

#endif
