/*
     This file is part of GNUnet.
     (C) 2001, 2002 Christian Grothoff (and other contributing authors)

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * This file contains the connection table which lists all the
 * current connections of the node with other hosts and buffers
 * outgoing packets to these hosts. The connection table also
 * contains state information such as sessionkeys, credibility and
 * the last time we had host activity.
 *
 * Together with sessionkey.c (and to a lesser extend pingpong.c and
 * heloexchange.c) this code is responsible for exchanging a sessionkey
 * with another peer, grouping several messages into a larger packet,
 * padding with noise, encryption and deferred sending of these messages.
 * <p>
 *
 * <em>A Note on synchronization and collision handling.</em>
 * The connection table is implemented as a hashtable. If a collision
 * occurs, a linked list is used. These linked lists currently NEVER
 * shrink since there is no good (always safe) way to shink them
 * without a global lock which is the last thing that we would want.
 * If you think you can do better, hack it up. But note that the list
 * may be grown at any time (the root-node is locked when we grow, though)
 * and that other threads may hold some locks in the linked list (including
 * the root-lock, but also just any other subset of locks is possible). 
 * And make sure we don't deadlock...<br>
 * 
 * Simple strategies do not work since you must aquire multiple locks
 * (all in the chain) to perform the unlinking, but there is no safe order
 * of aquiration since other threads may have aquired individual locks 
 * in no particular order. I'm not aware of a pthread-mechanism to aquire
 * multiple locks in one atomic step (or block as long as this is not 
 * possible). Note that we would also not want to block indefinitely
 * either.
 * <p>
 *
 * @author Tzvetan Horozov
 * @author Christian Grothoff
 **/ 

#include "config.h"
#include "util/statistics.h"
#include "util/logging.h"
#include "util/tcp_return.h"
#include "server/connection.h"
#include "server/knownhosts.h"
#include "server/trustservice.h"
#include "server/heloexchange.h"
#include "server/pingpong.h"

/**
 * Type of the linked list of send callbacks (to
 * implement a round-robbin invocation chain).
 **/
typedef struct SendCallbackList__ {
  /**
   * Minimum number of bytes that must be available
   * to call this callback.
   **/
  int minimumPadding;

  /**
   * The callback method.
   **/
  BufferFillCallback callback;

  /**
   * Did we say that this is a linked list?
   **/
  struct SendCallbackList__ * next;

} SendCallbackList;

/**
 * The buffer containing all current connections.
 **/
static BufferEntry * CONNECTION_buffer_;

/**
 * Size of the CONNECTION_buffer_
 **/
static unsigned int CONNECTION_MAX_HOSTS_;

/**
 * The DirContents array for scanning the hosts/ directory.
 **/
static int CONNECTION_currentActiveHosts;

/**
 * Experimental configuration: disable random padding of encrypted
 * messages.
 **/
static int disable_random_padding = NO;

/**
 * Send callbacks for making better use of noise padding...
 **/
static SendCallbackList * scl_nextHead;
static SendCallbackList * scl_nextTail;
static Mutex scl_mutex;

/**
 * statistics handles
 **/
static int stat_number_of_connections;
static int stat_number_of_bytes_noise_send;
static int stat_number_of_bytes_send;
static int stat_number_of_bytes_received;

/* **************** tuning parameters ************ */

/**
 * If an attempt to establish a connection is not answered
 * within 60s, drop.
 **/
#define SECONDS_NOPINGPONG_DROP 60

/**
 * If an established connection is inactive for 5 minutes, 
 * drop.
 **/
#define SECONDS_INACTIVE_DROP 300

/**
 * After 4 minutes on an inactive connection, probe the other
 * node with a ping if we have achieved less than 50% of our
 * connectivity goal.
 **/
#define SECONDS_PINGATTEMPT 240

/* ******************** CODE ********************* */

/**
 * From time to time, do a recount on how many hosts are connected.
 **/
static void cronCountConnections() {
  int act;
  int i;
  BufferEntry * root;
  BufferEntry * tmp;

  act = 0;
  for (i=0;i< CONNECTION_MAX_HOSTS_;i++) {
    root=&CONNECTION_buffer_[i];
    MUTEX_LOCK(&root->sem);

    tmp = root;
    while(NULL != tmp) {
      if (tmp->status == STAT_UP)
        act++;
      tmp = tmp->overflowChain;
    }
    
    MUTEX_UNLOCK(&root->sem);
  }

  CONNECTION_currentActiveHosts = act;
  statSet(stat_number_of_connections,
	  act);
}

/** 
 * This initializes a given, preallocated BufferEntry structure.
 **/
static void initBufferEntry(BufferEntry * be) {
    if(be == NULL)
      return;    
    MUTEX_CREATE_RECURSIVE(&be->sem);    
    be->trust 
      = 0;
    be->isAlive 
      = 0;
    be->status 
      = STAT_DOWN;
    be->prio 
      = 0;
    be->buffer
      = NULL;
    be->len 
      = sizeof(SEQUENCE_Message);
    be->overflowChain
      = NULL;
    be->session.tsession
      = NULL;
}

/**
 * Record for state maintanance between scanHelperCount,
 * scanHelperSelect and scanForHosts.
 **/
typedef struct {
  int index;
  int matchCount;
  int costSelector;  
  HostIdentity match;
} IndexMatch;

/**
 * Here in this scanning for applicable hosts, we also want to take
 * the protocols into account and prefer "cheap" protocols,
 * i.e. protocols with a low overhead.
 **/
static void scanHelperCount(const HostIdentity * id,
			    const unsigned short proto,	
			    IndexMatch * im) {
  if (hostIdentityEquals(&myIdentity, id)) 
    return;
  if (computeIndex(id) != im->index)
    return;
  if (YES == isTransportAvailable(proto)) {
    im->matchCount++;  
    im->costSelector += transportGetCost(proto);
  } else
    LOG(LOG_DEBUG,
	"DEBUG: transport %d is not available\n",
	proto);
}

static void scanHelperSelect(const HostIdentity * id,
			     const unsigned short proto,
			     IndexMatch * im) {
  if (hostIdentityEquals(&myIdentity, id)) 
    return;
  if (computeIndex(id) != im->index)
    return;
  if (YES == isTransportAvailable(proto)) {
    im->costSelector -= transportGetCost(proto);
    if ( (im->matchCount == 0) ||
	 (im->costSelector < 0) ) {
      memcpy(&im->match,
	     id,
	     sizeof(HostIdentity));
    }
    im->matchCount--;
  } else
    LOG(LOG_DEBUG,
	"DEBUG: transport %d is not available\n",
	proto);

}

/**
 * Look in the list for known hosts; pick a random host
 * of minimal transport cost for the hosttable at index 
 * index. When called, the 
 * mutex of at the given index must not be hold.
 **/
static void scanForHosts(unsigned int index) {
  BufferEntry * be;
  IndexMatch indexMatch;
  HexName hn;
  cron_t now;

  LOG(LOG_CRON, 
      "CRON: Scanning for hosts (%d)\n",
      index);
  cronTime(&now);
  indexMatch.index = index;
  indexMatch.matchCount = 0;
  indexMatch.costSelector = 0;
  forEachHost((HostIterator)&scanHelperCount, 
	      now,
	      &indexMatch);
  if (indexMatch.matchCount == 0) 
    return;  
  LOG(LOG_CRON, 
      "CRON: Scanning for hosts (%d) found %d matching node identities.\n",
      index,
      indexMatch.matchCount);    
  /* indexMatch.matchCount 
    = randomi(indexMatch.matchCount); randomly choose host */
  indexMatch.costSelector
    = randomi(indexMatch.costSelector);
  memcpy(&indexMatch.match,
	 &myIdentity,
	 sizeof(HostIdentity));  
  forEachHost((HostIterator)&scanHelperSelect,
	      now,
	      &indexMatch);
  if (hostIdentityEquals(&myIdentity, 
			 &indexMatch.match)) {
    LOG(LOG_ERROR, 
	"ERROR: scanHelperSelect did not find selected match!?\n");
    return;
  }
  if (computeIndex(&indexMatch.match) != index) {
    LOG(LOG_ERROR, 
	"ERROR: index of host to add is wrong!\n");
    return;
  }
  IFLOG(LOG_DEBUG,
	hash2hex(&indexMatch.match.hashPubKey,
		 &hn));
  LOG(LOG_DEBUG, 
      "DEBUG: attempting to connect to %s using slot %d.\n",
      &hn, 
      index);
  be = addHost(&indexMatch.match, 
	       NO); /* NO or YES should not matter here since we know the slot is empty */
  if (be != NULL) {
    if (be->status == STAT_DOWN) {
      blacklistHost(&be->session.sender,
		    CONNECTION_currentActiveHosts); 
      /* we're trying now, don't try again too soon */
      exchangeKey(be);
    }
    MUTEX_UNLOCK(&be->sem);
  }
}

/**
 * Consider switching the transport mechanism used for contacting
 * the given node. This function is called when the handler handles
 * an encrypted connection. For example, if we are sending SMTP
 * messages to a node behind a NAT box, but that node has established
 * a TCP connection to us, it might just be better to send replies
 * on that TCP connection instead of keeping SMTP going.<p>
 *
 * We can only successfuly takeover if the transport is bidirectional
 * (can be associated). It also only makes sense if the cost is lower.
 * This method checks both.
 *
 * @param tsession the transport session that is for grabs
 * @param sender the identity of the other node
 **/
void considerTakeover(TSession * tsession,
		      HostIdentity * sender) {
  BufferEntry * be;

  if (tsession == NULL)
    return;
  be = lookForHost(sender);
  if (be != NULL) {
    if (be->status != STAT_DOWN) {
      unsigned int cost = -1;
      if (be->session.tsession != NULL)
	cost = transportGetCost(be->session.tsession->ttype);
      if (transportGetCost(tsession->ttype) >= cost) {
	if (transportAssociate(tsession) == OK) {
	  transportDisconnect(be->session.tsession);
	  be->session.tsession = tsession;
	}
      } /* end if cheaper */
    } /* end if connected */
    MUTEX_UNLOCK(&be->sem);
  }
}

/**
 * Call this method periodically to decrease liveness of hosts.
 **/
static void cronDecreaseLiveness(void * unused) {  
  static int lastLivenessHost = 0;
  BufferEntry * root = NULL;
  BufferEntry * tmp;
  cron_t now;
  
  /* Find the correct host */
  root = &CONNECTION_buffer_[lastLivenessHost];
  cronTime(&now);    
  MUTEX_LOCK(&root->sem);
  tmp = root;
  while (NULL != tmp) {
    /* is this the host? */
    if (tmp != root)
      MUTEX_LOCK(&tmp->sem);
    switch (tmp->status) {
    case STAT_DOWN:
      if(tmp == root) {
	/* check that all entries in the
	   overflow chain are DOWN, if yes,
	   attempt to connect, if NO, just continue */
	while (tmp != NULL) {
	  if (tmp->status == STAT_DOWN)
	    tmp = tmp->overflowChain;
	  else
	    break; 
	} 
	if (tmp == NULL)
	  scanForHosts(lastLivenessHost);         	  
	tmp = root; 
      }
      break;
    case STAT_UP:
      if (now - tmp->isAlive > SECONDS_INACTIVE_DROP * cronSECONDS) {
	/* switch state form UP to DOWN: too much inactivity */
	shutdownConnection(tmp);
	whitelistHost(&tmp->session.sender); /* the host may still be worth trying again soon */
	CONNECTION_currentActiveHosts--;
	statChange(stat_number_of_connections,
		   -1);
      } else if ( (CONNECTION_currentActiveHosts < CONNECTION_MAX_HOSTS_/2) &&
		  (now - tmp->isAlive > SECONDS_PINGATTEMPT * cronSECONDS) ){
	/* if we have less than 50% of the number of connections
	   that we would like to have, try ping-ing the other side
	   to keep the connection open instead of hanging up */
	PINGPONG_Message pmsg;
	HostIdentity * hi = xmalloc(sizeof(HostIdentity), "hi clone");
	memcpy(hi, 
	       &tmp->session.sender, 
	       sizeof(HostIdentity));
	if (OK == pingAction(&tmp->session.sender,
			     (CronJob)&notifyPING,
			     hi,
			     &pmsg)) {
	  appendToBuffer(tmp, 
			 &pmsg.header, 
			 (SECONDS_INACTIVE_DROP - SECONDS_PINGATTEMPT) * cronSECONDS / 2);
	} else
	  xfree(hi, 
		"pingpong failure");
      }
      break;
    case STAT_WAITING_FOR_PING:
    case STAT_WAITING_FOR_PONG:
      if (now - tmp->isAlive > SECONDS_NOPINGPONG_DROP * cronSECONDS) 
	shutdownConnection(tmp);      
      break;
    default:
      LOG(LOG_WARNING,
	  "WARNING: unknown type of connection status: %d\n",
	  tmp->status);
      break; /* do nothing */
    } /* end of switch */ 
    if(tmp != root)
      MUTEX_UNLOCK(&tmp->sem);
    tmp = tmp->overflowChain;
  } /* end of while */
  MUTEX_UNLOCK(&root->sem);
  
  lastLivenessHost++;
  if(lastLivenessHost >= CONNECTION_MAX_HOSTS_)
    lastLivenessHost = 0;
}

/**
 * Check the sequence number. Updates the sequence
 * number as a side-effect.
 * @returns OK if ok, SYSERR if not.
 **/
static int checkSequenceNumber(HostIdentity * sender,
			       p2p_HEADER * msg) {
  SEQUENCE_Message * smsg;
  BufferEntry * be;
  int res;
  unsigned int sequenceNumber;
  
  if (ntohs(msg->size) != sizeof(SEQUENCE_Message)) {
    LOG(LOG_WARNING,
	"WARNING: sequence message received has wrong size %d\n",
	ntohs(msg->size));
    return SYSERR;
  }
  smsg = (SEQUENCE_Message*) msg;
  sequenceNumber = ntohl(smsg->sequenceNumber);
  be = lookForHost(sender);
  if (be == NULL) {
    LOG(LOG_FAILURE,
	"FAILURE: sequence message received for node that is not connected!?\n");
    return SYSERR; /* host not found */
  }
  res = OK;
  if (be->lastSequenceNumberReceived >= sequenceNumber) {
    unsigned int rotbit = 1;
    if ( (be->lastSequenceNumberReceived - sequenceNumber <= 32) && 
	 (be->lastSequenceNumberReceived != sequenceNumber) ) {
      rotbit = rotbit << (be->lastSequenceNumberReceived - sequenceNumber - 1);
      if ( (be->lastPacketsBitmap & rotbit) == 0) {
	res = OK;
	be->lastPacketsBitmap |= rotbit;
      } else
	res = SYSERR;
    } else
      res = SYSERR;
    if (res == SYSERR) {
      LOG(LOG_WARNING,
	  "WARNING: Invalid sequence number %u < %u, dropping rest of packet\n",
	  sequenceNumber, 
	  be->lastSequenceNumberReceived);
    }    
  } else {    
    be->lastPacketsBitmap = 
      be->lastPacketsBitmap << (sequenceNumber - be->lastSequenceNumberReceived);
    be->lastSequenceNumberReceived = sequenceNumber;
  }
  MUTEX_UNLOCK(&be->sem);
  return res;
}

/**
 * Check the sequence number. Updates the sequence
 * number as a side-effect.
 * @returns OK if ok, SYSERR if not.
 **/
static int processGetConnectionCountRequest(GNUNET_TCP_SOCKET * client,
					    CS_HEADER * msg) {
  if (ntohs(msg->size) != sizeof(CS_HEADER)) {
    LOG(LOG_WARNING,
	"WARNING: request received from TCP has wrong size %d\n",
	ntohs(msg->size));
    return SYSERR;
  }
  return sendTCPResult(client,
		       CONNECTION_currentActiveHosts);
}


/**
 * Handler for processing HANGUP.
 **/
static int handleHANGUP(HostIdentity * sender,
			p2p_HEADER * msg) {
  BufferEntry * be;
  HexName hex;

  if (ntohs(msg->size) != sizeof(HANGUP_Message))
    return SYSERR;
  if (!hostIdentityEquals(sender,
			  &((HANGUP_Message*)msg)->sender))
    return SYSERR;
  IFLOG(LOG_INFO,
	hash2hex(&sender->hashPubKey, 
		 &hex));
  LOG(LOG_INFO,
      "INFO: received HANGUP from %s\n",
      &hex);
  be = lookForHost(sender);
  if (be == NULL)
    return SYSERR;
  shutdownConnection(be);
  MUTEX_UNLOCK(&be->sem);
  return OK;
}


/**
 * Initialize this module.
 **/
void initConnection() {
  unsigned int i;

  stat_number_of_connections 
    = statHandle("# currently connected nodes");
  stat_number_of_bytes_noise_send
    = statHandle("# bytes noise sent");
  stat_number_of_bytes_send
    = statHandle("# encrypted bytes sent");
  stat_number_of_bytes_received
    = statHandle("# bytes decrypted");
  scl_nextHead 
    = NULL;
  scl_nextTail 
    = NULL;
  MUTEX_CREATE(&scl_mutex);
  CONNECTION_currentActiveHosts 
    = 0;
  CONNECTION_MAX_HOSTS_ 
    = getConfigurationInt("GNUNETD",
			  "MAXCONNECT");
  if (CONNECTION_MAX_HOSTS_ < 4)
    CONNECTION_MAX_HOSTS_ = 4;
   i = 1;
  while (i< CONNECTION_MAX_HOSTS_)
    i*=2;
  CONNECTION_MAX_HOSTS_ = i; /* make sure it's a power of 2 */

  CONNECTION_buffer_ 
    = (BufferEntry*) xmalloc(sizeof(BufferEntry)*CONNECTION_MAX_HOSTS_,
			     "initConnection: connection buffer");
  for (i=0;i<CONNECTION_MAX_HOSTS_;i++) 
    initBufferEntry(&CONNECTION_buffer_[i]); 

  registerp2pHandler(p2p_PROTO_SEQUENCE,
		     &checkSequenceNumber);
  registerp2pHandler(p2p_PROTO_HANGUP,
		     &handleHANGUP);
  registerCSHandler(CS_PROTO_CLIENT_COUNT,
		    &processGetConnectionCountRequest);
  addCronJob(&cronCountConnections,
	     1 * cronMINUTES,
	     30 * cronSECONDS,
	     NULL);
  addCronJob(&cronDecreaseLiveness,
	     1 * cronSECONDS, 
	     cronMINUTES / CONNECTION_MAX_HOSTS_,
	     NULL); /* Liveness in 1 minute intervals */  
  disable_random_padding = testConfigurationString("GNUNETD-EXPERIMENTAL",
						   "PADDING",
						   "NO");
}

/**
 * Shutdown the connection module.
 **/
void doneConnection() {
  int i;
  BufferEntry * be;
  
  unregisterCSHandler(CS_PROTO_CLIENT_COUNT,
		      &processGetConnectionCountRequest);
  unregisterp2pHandler(p2p_PROTO_SEQUENCE,
		     &checkSequenceNumber);
  delCronJob(&cronCountConnections,
	     30 * cronSECONDS, 
	     NULL);
  delCronJob(&cronDecreaseLiveness,
	     cronMINUTES / CONNECTION_MAX_HOSTS_, 
	     NULL); 
  for (i=0;i<CONNECTION_MAX_HOSTS_;i++) {
    be = &CONNECTION_buffer_[i];

    while(be) {
      MUTEX_LOCK(&be->sem);
      shutdownConnection(be);
      flushHostCredit(be, NULL);    
      MUTEX_DESTROY(&be->sem);
      be = be->overflowChain;
    }
  } 
  xfree(CONNECTION_buffer_,
	"doneConnection: connection buffer");
  CONNECTION_buffer_ = NULL;
}

/**
 * Print the contents of the connection buffer. May
 * NOT be called from synchronized context!
 **/
void printConnectionBuffer() {
  int i;
  BufferEntry * root;
  BufferEntry * tmp;
  HexName hostName;
  HexName skey;
  unsigned short ttype;

  for (i=0;i<CONNECTION_MAX_HOSTS_;i++) {
    root = &CONNECTION_buffer_[i];
    MUTEX_LOCK(&root->sem);
    
    tmp = root;
    while(tmp) {
      if (tmp->status != STAT_DOWN) {
        IFLOG(LOG_MESSAGE,
  	      hash2hex(&tmp->session.sender.hashPubKey, 
  	  	       &hostName));
	IFLOG(LOG_MESSAGE,
	      hash2hex((HashCode160*) &tmp->skey,
		       &skey));
	ttype = 0;
	if (tmp->session.tsession != NULL)
	  ttype = tmp->session.tsession->ttype;
	LOG(LOG_MESSAGE,
  	    "CONNECTION-TABLE: %3d-%1d-%2d-%6u-%4ds (of %ds)-%4u: %20s-%16s\n",
	    i,
	    tmp->status,
	    ttype,
	    tmp->trust & 0x7FFFFFFF, 
	    (int) ((cronTime(NULL) - tmp->isAlive)/cronSECONDS), 
	    SECONDS_INACTIVE_DROP,
	    tmp->len,
	    &hostName,
	    &skey);
      }
      tmp = tmp->overflowChain;
    }
    MUTEX_UNLOCK(&root->sem);
  }
}

static void sendBufferSynchronized(BufferEntry * be) {
  MUTEX_LOCK(&be->sem);
  sendBuffer(be);
  MUTEX_UNLOCK(&be->sem);
}

/**
 * Register a callback method that should be invoked whenever a
 * message is about to be send that has more than minimumPadding bytes
 * left before maxing out the MTU. The callback method can then be
 * used to add additional content to the message (instead of the
 * random noise that is added by otherwise). Note that if the MTU is 0
 * (for streams), the callback method will always be called with
 * padding set to the maximum number of bytes left in the buffer
 * allocated for the send.
 *
 * @param minimumPadding how large must the padding be in order
 *   to call this method?
 * @param callback the method to invoke. The receiver is the
 *   receiver of the message, position is the reference to the
 *   first unused position in the buffer where GNUnet is building
 *   the message, padding is the number of bytes left in that buffer.
 *   The callback method must return the number of bytes written to
 *   that buffer (must be a positive number).
 * @return OK if the handler was registered, SYSERR on error
 **/
int registerSendCallback(const int minimumPadding,
			 BufferFillCallback callback) {
  SendCallbackList * scl;

  scl = xmalloc(sizeof(SendCallbackList),
		"SCL in registerSendCallback");
  scl->minimumPadding = minimumPadding;
  scl->callback = callback;
  scl->next = NULL;
  MUTEX_LOCK(&scl_mutex);
  if (scl_nextTail == NULL) {
    scl_nextHead = scl;
    scl_nextTail = scl;
  } else {
    scl_nextTail->next = scl;
    scl_nextTail = scl;
  }
  MUTEX_UNLOCK(&scl_mutex);
  return OK;  
}
 
/**
 * Unregister a handler that was registered with registerSendCallback.
 * @return OK if the handler was removed, SYSERR on error
 **/
int unregisterSendCallback(const int minimumPadding,
			   BufferFillCallback callback) {
  SendCallbackList * pos;
  SendCallbackList * prev;

  prev = NULL;
  MUTEX_LOCK(&scl_mutex);
  pos = scl_nextHead;
  while (pos != NULL) {
    if ( (pos->callback == callback) &&
	 (pos->minimumPadding == minimumPadding) ) {
      if (prev == NULL) 
	scl_nextHead = pos->next;
      else
	prev->next = pos->next;
      if (scl_nextTail == pos)
	scl_nextTail = prev;
      xfree(pos,
	    "unregister send callback frees scl");
      MUTEX_UNLOCK(&scl_mutex);
      return OK;     
    }
    pos = pos->next;
  }
  MUTEX_UNLOCK(&scl_mutex);
  return SYSERR;
}

/**
 * We received a sign of life from this host. 
 **/
void notifyPONG(const HostIdentity * hostId) {
  BufferEntry * be;
  HexName hex;

  be = lookForHost(hostId);
  if (be != NULL) {
    switch (be->status) {
    case STAT_DOWN:
      break;
    case STAT_WAITING_FOR_PING:
      /* wrong message, ignore */
      break;
    case STAT_WAITING_FOR_PONG:
      be->status = STAT_UP;
      CONNECTION_currentActiveHosts++;
      cronTime(&be->isAlive);
      statChange(stat_number_of_connections,
		 1);
      IFLOG(LOG_DEBUG,
	    hash2hex(&hostId->hashPubKey, 
		     &hex));
      LOG(LOG_DEBUG, 
	  "DEBUG: Marking host %s active.\n",
	  &hex);
      break;
    case STAT_UP:
      cronTime(&be->isAlive);
      break;
    default:
      LOG(LOG_WARNING, 
	  "WARNING: undefined status (%d)!\n",
	  be->status);
      break;
    }
    MUTEX_UNLOCK(&be->sem);
  }
}

/**
 * We received a sign of life from this host. 
 **/
void notifyPING(const HostIdentity * hostId) {
  BufferEntry * be;
  HexName hex;

  be = lookForHost(hostId);
  if (be != NULL) {
    switch (be->status) {
    case STAT_DOWN:
      break;
    case STAT_WAITING_FOR_PONG:
      /* wrong message */
      break;
    case STAT_WAITING_FOR_PING:
      be->status = STAT_UP;
      CONNECTION_currentActiveHosts++;
      statChange(stat_number_of_connections,
		 1);
      cronTime(&be->isAlive);
      IFLOG(LOG_DEBUG,
	    hash2hex(&hostId->hashPubKey, 
		     &hex));
      LOG(LOG_DEBUG, 
	  "DEBUG: Marking host %s active.\n",
	  &hex);
      break;
    case STAT_UP:
      cronTime(&be->isAlive);
      break;
    default:
      LOG(LOG_WARNING, 
	  "WARNING: unknown status!\n");
      break;
    }
    MUTEX_UNLOCK(&be->sem);
  }
}

/**
 * Send a message to all directly connected nodes.
 *
 * @param message the message to send
 * @param priority how important is the message? The higher, the more important
 * @param maxdelay how long can we wait (max), in CRON-time (ms)
 **/
void broadcast(p2p_HEADER * message,
	       unsigned int priority,
	       unsigned int maxdelay) {
  unsigned int i;
  BufferEntry * root = NULL;
  BufferEntry * tmp;

  for (i=0;i<CONNECTION_MAX_HOSTS_;i++) {
    root = &CONNECTION_buffer_[i];
    MUTEX_LOCK(&root->sem);
    tmp = root;
    while (tmp != NULL) {
      /* we need no sync here as we only read,
         and concurrent rw access does not hurt */
      if (tmp->status == STAT_UP) 
	unicast(&tmp->session.sender,
		message,
		priority,
		maxdelay);
      tmp = tmp->overflowChain;
    }
    
    MUTEX_UNLOCK(&root->sem);    
  }
}

/**
 * Send a message to a specific host (reply, enqueue)
 *
 * @param message the message to send (unencrypted!), first 2 bytes give size
 * @param hostId the identity of the receiver
 * @param priority how important is the message?
 * @param maxdelay how long can we wait (max), in CRON-time (ms)
 **/
void unicast(const HostIdentity * hostId,
	     p2p_HEADER * message,
	     unsigned int priority,
	     unsigned int maxdelay) {
  BufferEntry * be;
  HexName hex;

  if (hostIdentityEquals(&myIdentity, 
			 hostId)) {
    LOG(LOG_ERROR, 
	"ERROR: attempt to unicast to myself! (dropped)\n");
    return;
  }
  IFLOG(LOG_DEBUG,
	hash2hex(&hostId->hashPubKey, 
		 &hex));
  LOG(LOG_DEBUG, 
      "DEBUG: unicasting to host %s message of type %d\n",
      &hex,
      ntohs(message->requestType));
  be = lookForHost(hostId);
  if ( (be == NULL) || 
       (be->status == STAT_DOWN) ) {
    if (be != NULL)
      MUTEX_UNLOCK(&be->sem);
    be = addHost(hostId, YES);   /* we *really* want to talk to this guy */
    if (be->status == STAT_DOWN)
      exchangeKey(be);
  }
  be->prio += priority;
  appendToBuffer(be,
		 message,
		 maxdelay);
  MUTEX_UNLOCK(&be->sem);
}

/**
 * Perform an operation for all connected hosts.  The BufferEntry
 * structure is passed to the method.  No synchronization or other
 * checks are performed.  This method relies on the fact that the
 * chains are never shortened or in an inconsistent state!
 *
 * @param method the method to invoke (NULL for couting only)
 * @param arg the second argument to the method
 * @return the number of connected hosts
 **/ 
int forAllConnectedHosts(BufferEntryCallback method,
			 void * arg) {
  unsigned int i;
  int count = 0;
  BufferEntry * be;
  
  for (i=0;i<CONNECTION_MAX_HOSTS_;i++) {
    be = &CONNECTION_buffer_[i];
    while (be != NULL) {
      if (be->status == STAT_UP) {
        if (method != NULL)
  	  method(be, arg);
        count++;
      }
      be = be->overflowChain;
    }
  }
  return count;
}
	       
/**
 * Decipher data coming in from a foreign host.
 *
 * @param data the data to decrypt
 * @param size the size of the encrypted data
 * @param hostId the sender host that encrypted the data 
 * @param result where to store the decrypted data, must
 *        be at least of size data->len long
 * @returns the size of the decrypted data, SYSERR on error
 **/
int decryptFromHost(void * data,
		    const unsigned short size,
		    const HostIdentity * hostId,
		    void * result) {  
  BufferEntry * be;
  int res;
  HexName hex;

  statChange(stat_number_of_bytes_received, size);
  if ( (data == NULL) || 
       (hostId == NULL) ) {    
    LOG(LOG_FAILURE, 
	"CONNECTION: could not decrypt, message or hostId was NULL\n");
    breakpoint();
    return SYSERR;
  }
  IFLOG(LOG_DEBUG,
	hash2hex(&hostId->hashPubKey, 
		 &hex));
  LOG(LOG_DEBUG, 
      "DEBUG: decrypting message from host %s\n",
      &hex);
  be = lookForHost(hostId);
  if (be == NULL) {
    IFLOG(LOG_FAILURE,
	  hash2hex(&hostId->hashPubKey, 
		   &hex));
    LOG(LOG_FAILURE, 
	"FAILURE: decrypting message from host %s failed, no sessionkey!\n",
	&hex);
    return SYSERR; /* could not decrypt */
  }
  res = decryptBlock(&be->skey, 
		     data,
		     size,
		     INITVALUE,
		     result);
  MUTEX_UNLOCK(&be->sem);
  return res;
}

/**
 * Look for a host in the table. If the entry is there
 * at the time of checking, returns the entry (the lock
 * is acquired).
 * @returns the index of the host in the table
 **/
BufferEntry * lookForHost(const HostIdentity * hostId) {
  unsigned int hash;
  BufferEntry * root;
  BufferEntry * tmp;

  hash = computeIndex(hostId);
  root = &CONNECTION_buffer_[hash];
  MUTEX_LOCK(&root->sem);
  if (equalsHashCode160(&hostId->hashPubKey,
			&root->session.sender.hashPubKey)) {
    return root;
  } else {
    /* check the overflowChain */
    tmp = root->overflowChain;

    while (tmp != NULL) {
      MUTEX_LOCK(&tmp->sem);
      if (equalsHashCode160(&hostId->hashPubKey,
			    &tmp->session.sender.hashPubKey)) {
	MUTEX_UNLOCK(&root->sem);
        return tmp;
      }
      MUTEX_UNLOCK(&tmp->sem);
      tmp = tmp->overflowChain;
    }
	
    MUTEX_UNLOCK(&root->sem);    
    return NULL; /* not found! */
  }
}

/**
 * Force adding of a host to the buffer. If the node is already in the
 * table, the table entry is returned.  
 *
 * @param force if YES, drop another host from the table if the slot
 * is already in use. If NO, return NULL if the slot is busy.  
 * @return the table entry for the host (no keyexchange performed so far)
 **/
BufferEntry * addHost(const HostIdentity * hostId,
		      int force) {
  BufferEntry * be;
  BufferEntry * root;
  BufferEntry * tmp;
  BufferEntry * prev;
  HexName hex;

  IFLOG(LOG_INFO,
	hash2hex(&hostId->hashPubKey, 
		 &hex));
  LOG(LOG_INFO, 
      "INFO: Adding host %s to the connection table.\n",
      &hex);

  be = lookForHost(hostId);
  if (be != NULL)
    return be;

  root = &CONNECTION_buffer_[computeIndex(hostId)];
  MUTEX_LOCK(&root->sem);
    
  /* if root is not down, look from overflow chain */
  if(root->status != STAT_DOWN) {
    prev = root;
    tmp = root->overflowChain;
    while (NULL != tmp) {
      MUTEX_LOCK(&tmp->sem);

      /* settle for entry in the linked list that is down */
      if(tmp->status == STAT_DOWN) {
	be = tmp;
	break;
      }
      MUTEX_UNLOCK(&tmp->sem);
      tmp = tmp->overflowChain;
      prev = tmp;
    }
      
    /** if we didn't find suitable, allocate new **/
    if(be == NULL) {
      be = xmalloc(sizeof(BufferEntry),
		   "addHost: adding to overflowChain");
      initBufferEntry(be);
      prev->overflowChain = be;
      /* lock new */
      MUTEX_LOCK(&prev->overflowChain->sem);
    }
    /* release root lock */
    MUTEX_UNLOCK(&root->sem);
  } else
    be = root; /* else use root */ 

  /* overwrite/create new entry */
  memcpy(&be->session.sender,
	 hostId,
	 sizeof(HostIdentity));	 
  initHostTrust(be);
  return be;
}

/**
 * Compute the hashtable index of a host id.
 **/
unsigned int computeIndex(const HostIdentity * hostId) {
  unsigned int res = (((unsigned int)hostId->hashPubKey.a) & 
		      ((unsigned int)(CONNECTION_MAX_HOSTS_ - 1)));
  if (res >= CONNECTION_MAX_HOSTS_)
    errexit("FATAL: CONNECTION_MAX_HOSTS_ not power of 2? (%d)\n",
	    CONNECTION_MAX_HOSTS_);
  return res;
}

/**
 * Append a message to the current buffer. This method
 * assumes that the access to be is already synchronized.
 **/
void appendToBuffer(BufferEntry * be,
		    p2p_HEADER * message,
		    unsigned int maxdelay) {
  unsigned short exSize;
  unsigned short newSize;
  HexName hex;
  
  IFLOG(LOG_DEBUG,
	hash2hex(&be->session.sender.hashPubKey, 
		 &hex));
  LOG(LOG_DEBUG, 
      "DEBUG: adding message of size %d and type %d to buffer of host %s.\n",
      ntohs(message->size),
      ntohs(message->requestType),
      &hex);
  if (message == NULL) {
    LOG(LOG_WARNING, 
	"WARNING: appendToBuffer got garbage. Ignored.\n");
    return;
  }
  if (be->buffer == NULL) {
    LOG(LOG_FAILURE,
	"FAILURE: be->buffer NULL in append to buffer. Dropping message.\n");
    breakpoint();
    return;
  }

  newSize = ntohs(message->size);
  exSize = be->len;
  if (exSize + newSize > be->session.mtu) {
    if (newSize > be->session.mtu) {
      LOG(LOG_FAILURE, 
	  "FAILURE: appendToBuffer fatal error: exSize: %d newsize: %d BUFSIZE %d\n",
	  exSize, 
	  newSize, 
	  be->session.mtu);
      breakpoint();
      return;
    }
    sendBuffer(be); /* clean up*/
    exSize = sizeof(SEQUENCE_Message);
    if (exSize + newSize > be->session.mtu) {
      LOG(LOG_FAILURE, 
	  "FAILURE: appendToBuffer fatal error: exSize: %d newsize: %d BUFSIZE %d\n",
	  exSize, 
	  newSize, 
	  be->session.mtu);
      breakpoint();
      return;
    }
  }
  memcpy(&((char*)be->buffer)[exSize], 	 
	 message, newSize); 
  be->len = exSize + newSize;

  /* flush the buffer after an average of maxdelay/2 seconds;
     if a message is already pending, do not schedule another
     cron job *except* if it's really urgent (instant send) */
  if (maxdelay == 0) {
    if (be->flushTime != 0)
      delCronJob((CronJob)sendBufferSynchronized,
		 0,
		 be);
    sendBuffer(be);
  } else {
    cron_t now;
    int delta;

    delta = randomi(maxdelay+1);
    cronTime(&now);
    LOG(LOG_DEBUG,
	"DEBUG: sending buffer deferred for at most %dms (%u)\n",
	delta,
	maxdelay);
    if (be->flushTime == 0) {
      be->flushTime = delta + now;
      addCronJob((CronJob)sendBufferSynchronized,
		 delta,
		 0,
		 be);
    } else if (be->flushTime > delta + now) {
      delCronJob((CronJob)sendBufferSynchronized,
		 0, 
		 be);
      be->flushTime = delta + now;
      LOG(LOG_DEBUG,
	  "DEBUG: sending buffer deferred in at most %dms\n",
	  delta);
      addCronJob((CronJob)sendBufferSynchronized,
		 delta,
		 0,
		 be);     
    }
  }
}

/**
 * Send a buffer; assumes that access to be is already 
 * synchronized.
 **/
void sendBuffer(BufferEntry * be){
  int crc;
  char * content;
  p2p_HEADER * part;
  int i;
  SendCallbackList * pos;

  if (be == NULL) {
    LOG(LOG_WARNING, 
	"WARNING: sendBuffer called with invalid arguments (NULL).\n");
    return;
  }  
  if (be->status == STAT_DOWN)
    return; /* status is down, nothing to send! (should at least be wait-for-ping/pong 
	       or up) */
  if (be->buffer == NULL) {
    LOG(LOG_FAILURE,
	"FAILURE: assertion failed: buffer NULL but status not down (%s:%d)\n",
	__FILE__, __LINE__);
    return; /* fail gracefully... */
  }
  be->flushTime = 0; /* flushing now! */
  if (be->len <= sizeof(SEQUENCE_Message))       
    return; /* nothing to send */  
  if (SYSERR == outgoingCheck(be->prio)) {
    be->prio 
      = 0;
    be->len 
      = sizeof(SEQUENCE_Message); /* too busy, truncate and drop! */
    LOG(LOG_DEBUG,
	"DEBUG: policy (outgoing check) dropped message\n");
    return;
  }
  /* try callbacks */
  MUTEX_LOCK(&scl_mutex);
  pos = scl_nextHead;
  while (pos != NULL) {
    if (pos->minimumPadding + be->len <= be->session.mtu) 
      be->len += pos->callback(&be->session.sender,
			       &((char*)be->buffer)[be->len],
			       be->session.mtu - be->len);
    pos = pos->next;
  }
  MUTEX_UNLOCK(&scl_mutex);

  /* finally padd with noise */  

  if ( (be->len + sizeof(p2p_HEADER) <= be->session.mtu) &&
       (disable_random_padding == NO) ) {
    unsigned short noiseLen = be->session.mtu - be->len;

    LOG(LOG_DEBUG,
	"DEBUG: padding with %d bytes of noise at %d\n",
	noiseLen, be->len);
    statChange(stat_number_of_bytes_noise_send,
	       noiseLen);
    part = (p2p_HEADER *) &((char*)be->buffer)[be->len];
    part->size 
      = htons(noiseLen);
    part->requestType 
      = htons(p2p_PROTO_NOISE);
    for (i=be->len+sizeof(p2p_HEADER)-sizeof(SEQUENCE_Message);
	 i < be->session.mtu-sizeof(SEQUENCE_Message);
	 i++)
      be->buffer->variable[i] = (char) rand();
    be->len 
      = be->session.mtu;
  }

  crc = crc32N(be->buffer, 
	       be->len);
  content = xmalloc(be->len,
		    "alloc buffer for outgoing message");
  if (be->len == encryptBlock(be->buffer,
			      be->len,
			      &be->skey,
			      INITVALUE,
			      content)) {    
    statChange(stat_number_of_bytes_send,
	       be->len);
    LOG(LOG_DEBUG,
	"DEBUG: calling transport layer to send %d bytes\n",
	be->len);
    transportSend(be->session.tsession,
		  content,
		  be->len,
		  YES, /* is encrypted */
		  crc);  
  } else
    LOG(LOG_FAILURE,
	"FAILURE: Encryption of message failed!\n");
  xfree(content,
	"message is out");
  be->prio 
    = 0;
  be->len 
    = sizeof(SEQUENCE_Message); /* ok, mark buffer as send */
  be->buffer->sequence.sequenceNumber 
    = htonl(ntohl(be->buffer->sequence.sequenceNumber)+1);
}

/* end of connection.c */
