/*
     This file is part of GNUnet

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/
/**
 * Main handler for incoming packets.
 *
 * @author Christian Grothoff
 * @file server/handler.c
 **/

#include "config.h"
#include "server/handler.h"
#include "server/policy.h"
#include "server/sessionkey.h"
#include "server/pingpong.h"
#include "server/heloexchange.h"
#include "server/connection.h"
#include "util/semaphore.h"
#include "util/statistics.h"

/**
 * Array of the message handlers.
 **/
static MessagePartHandler * handlers = NULL;

/**
 * Number of handlers in the array (max, there
 * may be NULL pointers in it!)
 **/
static unsigned short max_registeredType = 0;

/**
 * Mutex to guard access to the handler array.
 **/
static Mutex handlerLock;

/**
 * Noise received-stats.
 **/
static int stat_bytes_noise_received;

/**
 * Handler for processing noise.
 **/
static int processNoise(HostIdentity * sender,
			p2p_HEADER * msg) {
  statChange(stat_bytes_noise_received,
	     ntohs(msg->size));
  return OK;
}

/**
 * Initialize message handling module.
 **/
void initHandler() {
  int i;

  MUTEX_CREATE(&handlerLock);
  max_registeredType = 32;
  handlers = xmalloc(max_registeredType *
		     sizeof(MessagePartHandler),
		     "handler array (default size)");
  for (i=0;i<max_registeredType;i++)
    handlers[i] = NULL;
  stat_bytes_noise_received 
    = statHandle("# bytes of noise received");
  if (SYSERR == registerp2pHandler(p2p_PROTO_NOISE,
				   &processNoise))
    errexit("FATAL: could not registerp2p handler for noise!?\n");
}

/**
 * Shutdown message handling module.
 **/
void doneHandler() {
  MUTEX_DESTROY(&handlerLock);
  xfree(handlers,
	"freeing handler array");
  max_registeredType = 0;
  handlers = NULL;
}

/**
 * Handle a message (that was decrypted if needed).
 * Checks the CRC and if that's ok, processes the
 * message by calling the registered handler for
 * each message part.
 **/
static void handleHelper(char * msg,
			 HostIdentity * sender,
			 const unsigned int size,
			 const int crc) {
  unsigned int pos;
  p2p_HEADER * part;

  if (crc32N(msg, size) != crc) {
    LOG(LOG_WARNING, 
	"WARNING: CRC check failed, message ignored (wrong sessionkey?)\n");
    return;
  }
  pos = 0;
  while (pos < size) {
    unsigned short plen;
    unsigned short ptyp;
    MessagePartHandler callback;

    part = (p2p_HEADER *) &msg[pos];
    plen = htons(part->size);
    if (pos + plen > size) {
      LOG(LOG_WARNING, 
	  "WARNING: handleDecrypted: message corrupted (%d+%d > %d)!\n",
	  pos, plen, size);
      break; /* exit */
    }
    pos += plen;

    ptyp = htons(part->requestType);
    if (ptyp > max_registeredType) {
      LOG(LOG_INFO, 
	  "INFO: handleDecrypted: Message not understood: %d (no handler registered)\n",
	   ptyp);
      continue; /* no handler registered, go to next part */
    }

    callback = handlers[ptyp];
    if (callback == NULL) {
      LOG(LOG_INFO, 
	  "INFO: handleDecrypted: Message not understood: %d (no handler registered)\n",
	  ptyp);
      continue; /* no handler registered, go to next part */
    }
    if (SYSERR == callback(sender,
			   part))
      break; /* handler says: do not process the rest of
		the message */
  }
}

/**
 * Handle a message (that was decrypted if needed).
 * Checks the CRC and if that's ok, processes the
 * message by calling the registered handler for
 * each message part.
 **/
static void handlePlaintext(char * msg,
			    TSession * tsession,
			    HostIdentity * sender,
			    const unsigned int size,
			    const int crc) {
  unsigned int pos;
  p2p_HEADER * part;

  if (crc32N(msg, size) != crc) {
    LOG(LOG_WARNING, 
	"WARNING: CRC check failed (%d expected, got %d), message of size %d ignored\n",
	crc, crc32N(msg, size), size);
    return;
  }
  pos = 0;
  while (pos < size) {
    unsigned short plen;
    unsigned short ptyp;

    part = (p2p_HEADER *) &msg[pos];
    plen = ntohs(part->size);
    if (pos + plen > size) {
      LOG(LOG_WARNING, 
	  "WARNING: handlePlaintext: message corrupted (pos(%d) + plen(%d) > size(%d))\n",
	  pos, plen, size);
      break; /* exit */
    }
    pos += plen;
    
    ptyp = htons(part->requestType);
    switch (ptyp) {
    case p2p_PROTO_HELO: 
      receivedHELO(part);
      break;    
    case p2p_PROTO_SKEY:
      /* establish session if slot not busy,
	 challenge with (encrypted) ping! */
      acceptSessionKey(sender,
		       tsession,
		       part);
      break;
    case p2p_PROTO_PING:
      /* challenge: send back reply - NOW! */
      plaintextPingReceived(sender, tsession, part);
      break;
    case p2p_PROTO_PONG:
      /* this confirms a PING => add a HELO to
	 knownhosts. */
      plaintextPongReceived(sender, tsession, part);
      break;
    default:
      LOG(LOG_WARNING, 
	  "WARNING: Message type %d not supported (for plaintext)\n",
	  ptyp);
      break;
    }
  }
}

/**
 * The actual main method of GNUnet: message dispatch/handling.
 * @param msg the message that was received. caller frees it on return
 **/
void handleMessage(TSession * tsession,
		   HostIdentity * sender,
		   void * msg,
		   const unsigned int size,
		   int isEncrypted,
		   const int crc) {
  /* check policy (is sender host blocked? is load ok?) */
  if (SYSERR == incomingCheck(sender)) {
    return; /* packet denied by policy */
  }
  if (isEncrypted == YES) {
    char * plaintext;
    
    plaintext = xmalloc(size,
			"space for plaintext message");
    if (SYSERR == decryptFromHost(msg, 
				  size,
				  sender,
				  plaintext)) {
      HexName hostName;
      
      IFLOG(LOG_WARNING,
	    hash2hex(&sender->hashPubKey, 
		     &hostName));
      LOG(LOG_WARNING, 
	  "WARNING: Can not decrypt packet from host %s: no session key? (%d).\n",
	  &hostName,
	  size);
      xfree(plaintext,
	    "decrypt failed, no plaintext produced");
      return;
    } 
    /* we may be able to use this transport-session to reduce our
       cost to send replies to the sender, let's check */
    considerTakeover(tsession, sender);
    handleHelper(plaintext, sender, size, crc);
    xfree(plaintext,
	  "handler done, freeing plaintext message");
  } else {
    handlePlaintext(msg, tsession, sender, size, crc);
  }
}

/**
 * Register a method as a handler for specific message
 * types.
 * @param type the message type
 * @param callback the method to call if a message of
 *        that type is received
 * @return OK on success, SYSERR if there is already a
 *         handler for that type
 **/
int registerp2pHandler(const unsigned short type,
		       MessagePartHandler callback) {
  MUTEX_LOCK(&handlerLock);
  if (type < max_registeredType) {
    if (handlers[type] != NULL) {
      MUTEX_UNLOCK(&handlerLock);
      LOG(LOG_WARNING,
	  "WARNING: could not register handler for type %d (slot used)\n",
	  type);
      return SYSERR;
    } 
  } else {  /* must grow */
    int nSize;
    int i;
    MessagePartHandler * tmp;

    nSize = type + 32;
    tmp = xmalloc(nSize *
		  sizeof(MessagePartHandler),
		  "handler array (growing)");
    memcpy(tmp,
	   handlers,
	   max_registeredType *
	   sizeof(MessagePartHandler));
    xfree(handlers,
	  "freeing handler array (old, growing)");
    for (i = max_registeredType;i<nSize;i++)
      tmp[i] = NULL; /* zero out! */
    max_registeredType = nSize;
    handlers = tmp;
  }
  handlers[type] = callback;
  MUTEX_UNLOCK(&handlerLock);
  return OK;    
}
  
/**
 * Unregister a method as a handler for specific message
 * types. Only for encrypted messages!
 * @param type the message type
 * @param callback the method to call if a message of
 *        that type is received
 * @return OK on success, SYSERR if there is a different
 *         handler for that type
 **/
int unregisterp2pHandler(const unsigned short type,
			 MessagePartHandler callback) {
  MUTEX_LOCK(&handlerLock);
  if (type < max_registeredType) {
    if (handlers[type] != callback) {
      MUTEX_UNLOCK(&handlerLock);
      return SYSERR;
    } else {
      handlers[type] = NULL;
      MUTEX_UNLOCK(&handlerLock);
      return OK;
    }
  } 
  MUTEX_UNLOCK(&handlerLock);
  return SYSERR; 
}

/* end of handler.c */
