/*
     This file is part of GNUnet.
     (C) 2001, 2002 Christian Grothoff (and other contributing authors)

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * Policy implementation.
 * @author Christian Grothoff
 * @file server/policy.c
 **/

#include "config.h"
#include "server/policy.h"
#include "util/configuration.h"
#include "server/trustservice.h"

/**
 * Configuration...
 **/
static CIDRNetwork * trustedNetworks_ = NULL;

/**
 * Initialize the policy module.
 **/
void initPolicy() {
  char * ch;

  ch = getConfigurationString("NETWORK",
			      "TRUSTED");
  if (ch == NULL) {
    trustedNetworks_ = parseRoutes("127.0.0.0/8;"); /* by default, trust localhost only */
  } else {
    trustedNetworks_ = parseRoutes(ch);    
    if (trustedNetworks_ == NULL) 
      errexit("Malformed entry in the configuration in section %s under %s: %s\n",
	      "NETWORK",
	      "TRUSTED", 
	      ch); 
    xfree(ch,
	  "done parsing trusted networks string in initPolicy");
  }
}

/**
 * A new packet is supposed to be send out. Should it be
 * dropped because the load is too high?
 * <p>
 * @param priority the highest priority of contents in the packet
 * @return OK if the packet should be handled, SYSERR if the packet should be dropped.
 **/
int outgoingCheck(unsigned int priority) {
  int delta;

  delta = 100 - getNetworkLoad(); /* how much freedom do we have? */
  if (delta * delta * delta * (priority+1) < 50*50*50) {
    LOG(LOG_DEBUG, 
	    "DEBUG: network load too high (%d, priority is %u; %d < %d),"\
      "dropping outgoing.\n",
      100 - delta, priority, delta * delta * delta * (priority+1), 50*50*50);
    return SYSERR;
  } else
    LOG(LOG_DEBUG, 
	"DEBUG: network load ok (%d, priority is %u; %d >= %d), sending outgoing.\n",
	100 - delta, priority,
	delta * delta * delta * (priority+1), 50*50*50);
  return OK; /* no policy */
}

/**
 * A new packet is coming in. The question is, if we should
 * have a look at it at all (incl. decryption). The sender
 * is known. 
 * <p>
 * The policy must decide if we should drop the packet. The
 * reason may be that we don't trust the host or that the load
 * is to high (or both or a combination). 
 * <p>
 * @param senderIdentity the identity of the sender
 * @return OK if the packet should be handled, SYSERR if the packet should be dropped.
 **/
int incomingCheck(const HostIdentity * senderIdentity) {
  if (getCPULoad() > 100) {
    LOG(LOG_DEBUG, 
	"DEBUG: CPU load to high (%d), dropping incoming packet.\n",
	getCPULoad());
    return SYSERR;
  }
  return OK; /* no policy */
}

/**
 * Is this IP labeled as trusted for CS connections?
 **/
int isWhitelisted(struct in_addr ip) {
  return checkIPListed(trustedNetworks_,
		       ip);
}

/* end of policy.c */

